""" Module for managing exceptions raised by ERAD package."""

# standard libraries
from pathlib import Path


class ERADBaseException(Exception):
    """All exception should derive from this."""


class FeatureNotImplementedError(ERADBaseException):
    """Exception raised because specific feature requested has not been implemented."""


class PathDoesNotExist(ERADBaseException):
    """Exception raised bacause expected file/folder path does not exist."""

    def __init__(self, path):
        self.message = (
            f"Expected path {path} does not exist. please check you file path!"
        )
        super().__init__(self.message)


class NotAFileError(ERADBaseException):
    """Exception raised because file is expected but folder path is provided."""

    def __init__(self, path):
        self.message = f"Expected file path {path} is not a file!"
        super().__init__(self.message)


class EmptyEnvironmentVariable(ERADBaseException):
    """Exception raised because environment variable required is empty."""


class DatabaseMissingInfo(ERADBaseException):
    """Exception raised because information required to connect to database is missing."""


class InvalidFileTypePassed(ERADBaseException):
    """Exceptions raised because invalid file type is passed."""

    def __init__(self, path, valid_type):

        self.message = f"Invalid file type of {Path(path).suffix} is passed! Please pass valid file type of {valid_type}"
        super().__init__(self.message)


class SMARTDSInvalidInput(ERADBaseException):
    """Exceptions raised because invalid input is provided for SMART DS data download."""


class EmptyScenarioPolygon(ERADBaseException):
    """Exceptions raised because no polygons are found."""


class OpenDSSCommandError(ERADBaseException):
    """Exceptions raised because opendss command execution ran into an error."""


class MultiStatePlaneError(ERADBaseException):
    """Exceptions raised because the corrdinates are in more than one state plane
    coordinates."""


class DittoException(ERADBaseException):
    """Exceptions raised because application ran into an issus using Ditto."""



""" Module for managing constants in ERAD package.

_Do not change this constants in your code._
"""

import os

ROOT_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DATA_FOLDER_NAME = "data"
DATA_FOLDER = os.path.join(ROOT_PATH, DATA_FOLDER_NAME)

FIRE_HISTORIC_GEODATAFRAME_PATH = "US_Wildfires_1878_2019.gdb"
EARTHQUAKE_HISTORIC_CSV_PATH = "World_Earthquakes_1960_2016.csv"
FLOOD_HISTORIC_SHP_PATH = "FEMA_100_Year_Flood_Zones_in_the_US\\FEMA_100_Year_Flood_Zones_in_the_US.shp"
#FLOOD_HISTORIC_SHP_PATH = "NFHL_06_20230323.gdb"
ELEVATION_RASTER_FILE = "TrueMarble.250m.21600x21600.B2.tif"

SMARTDS_VALID_YEARS = [2016, 2017, 2018]
SMARTDS_VALID_AREAS = ['SFO', 'GSO', 'AUS']




from erad.constants import FIRE_HISTORIC_GEODATAFRAME_PATH, DATA_FOLDER
from shapely.geometry import MultiPolygon, Point, LineString
from erad.scenarios.utilities import ProbabilityFunctionBuilder
from erad.scenarios.abstract_scenario import BaseScenario
from erad.exceptions import FeatureNotImplementedError
from erad.scenarios.utilities import GeoUtilities
import matplotlib.pyplot as plt
from datetime import datetime
import geopandas as gpd
import numpy as np
import random
import pyproj
import os

from erad.scenarios.common import AssetTypes
from erad.scenarios.utilities import ProbabilityFunctionBuilder


class WindScenario(BaseScenario, GeoUtilities): 
    """Base class for FireScenario. Extends BaseScenario and GeoUtilities

    Attributes:
        multipolygon (MultiPolygon): MultiPolygon enclosing wild fire regions 
        probability_model (dict): Dictionary mapping asset types to probability funcitons
        timestamp (datetime): Scenario occurance time 
    """
    
    fragility_curves = {
        #Extending energy system modelling to include extreme weather risks and application to hurricane events in Puerto Rico
        AssetTypes.substation.name : ProbabilityFunctionBuilder("lognorm", [0.8, 10, 5]),
        AssetTypes.solar_panels.name :  ProbabilityFunctionBuilder("lognorm", [0.8, 10, 5]),
        AssetTypes.buried_lines.name :  ProbabilityFunctionBuilder("lognorm", [0.8, 10, 5]),
        AssetTypes.wind_turbines.name :  ProbabilityFunctionBuilder("norm", [0.8, 10, 5]),
        #AssetTypes.battery_storage.name : ProbabilityFunctionBuilder("lognorm", [0.8, 10, 5]),
        #AssetTypes.transmission_poles.name  :   ProbabilityFunctionBuilder("lognorm", [0.8, 10, 5]),
        AssetTypes.distribution_poles.name  :  ProbabilityFunctionBuilder("lognorm", [0.8, 10, 5]),
       # AssetTypes.transmission_overhead_lines.name  : ProbabilityFunctionBuilder("lognorm", [0.8, 10, 5]),
        AssetTypes.distribution_overhead_lines.name  :  ProbabilityFunctionBuilder("beta", [0.8, 10, 5]),
    }
    
    def __init__(self,  multipolygon : MultiPolygon , probability_model : dict, timestamp : datetime) -> None:
        """Constructor for FireScenario.

        Args:
            multipolygon (MultiPolygon):  MultiPolygon enclosing wild fire regions 
            probability_model (dict): Dictionary mapping asset types to probability funcitons
            timestamp (datetime): Scenario occurance time 
        """

        super(WindScenario, self).__init__(multipolygon, probability_model, timestamp)     
        return      
    
    @property
    def area(self) -> float:
        """Method to calculate area of affected region."""
        geod = pyproj.Geod(ellps="WGS84")
        area = abs(geod.geometry_area_perimeter(self.polygon)[0])
        return area
    
    @property
    def polygon(self) -> MultiPolygon:
        """Method to return polygon for the affected region."""
        return self.multipolygon
        
    @property
    def boundary(self) -> LineString:
        """Method to return boundary for the affected region."""
        return self.multipolygon.boundary
        
    @property
    def centroid(self) -> Point:
        """Method to return the centroid of the affected region."""
        return self.polygon.centroid
            
    def increment_time(self):
        """Method to increment simulation time for time evolviong scenarios."""
        raise FeatureNotImplementedError()

    def calculate_survival_probability(self, assets : dict, timestamp : datetime, plot: bool) -> dict:
        """Method to calculate survival probaility of asset types.

        Args:
            assets (dict): The dictionary of all assets and their corresponding asset types
            plot (bool): Set to true to plot the fire survival model
        """     
        return assets
                





""" Utility functions that can be used in various parts of the code. """

# standard libraries
import json
from pathlib import Path
import logging
import logging.config
import time
from typing import Union

# third-party libraries
import yaml
import geojson

# internal imports
from erad.exceptions import (
    FeatureNotImplementedError,
    PathDoesNotExist,
    NotAFileError,
    InvalidFileTypePassed,
)

logger = logging.getLogger(__name__)


def timeit(func):
    """Decorator for timing execution of a function."""

    def wrapper(*args, **kwargs):
        time_start = time.perf_counter()
        logger.debug(f"Timing for {func} started")
        ret_val = func(*args, **kwargs)
        time_elapsed = time.perf_counter() - time_start
        # memory_mb =resource.getrusage(resource.RUSAGE_SELF).ru_maxrss/1024.0/1024.0
        logger.debug(
            f"Time took to execute the function {func} \
            with args {args}, kwargs {kwargs} is {time_elapsed} seconds"
        )
        return ret_val

    return wrapper


def setup_logging(filename: Union[str, None] = None) -> None:
    """Creates log directory and sets up logging via logging.yaml.

    Args:
        filename (str): Path to logging.yaml file

    If not providex expects log file in the root of repo.
    """

    if filename is None:
        filename = Path(__file__).parents[2] / "logging.yaml"

    logging.config.dictConfig(read_file(filename))


def path_validation(
    file_path: str,
    check_for_file: bool = False,
    check_for_file_type: Union[str, None] = None,
) -> None:
    """Utility function for validating the path.

    Args:
        file_path (str): Path to be validated
        check_for_file (bool): Checks for existence of file
        check_for_file_type (Union[str, None]): Check if file is of
            this type

    Raises:
        PathDoesNotExist: Raises if path does not exist
        NotAFileError: Raises if file is not present
        InvalidFileTypePassed: Raises if invalid file type is passed
    """

    file_path = Path(file_path)
    if not file_path.exists():
        logger.error(f"{file_path} does not exist!")
        raise PathDoesNotExist(file_path)

    if check_for_file and file_path.is_dir():
        logger.error(f"Expected file but got folder : {file_path} ")
        raise NotAFileError(file_path)

    if check_for_file_type and file_path.suffix != check_for_file_type:
        raise InvalidFileTypePassed(file_path, check_for_file_type)

    logger.debug(f"{file_path} validated successfully!")


@timeit
def read_file(file_path: str) -> dict:
    """Utility function to read file into a python dict.

    Supports json, yaml and geojson.

    Args:
        file_path (str): Path to a file to be read.

    Raises:
        FeatureNotImplementedError: Raises if invalid file is passed.

    Returns:
        dict: Python dict containing content of file.
    """

    file_path = Path(file_path)
    logger.debug(f"Attempting to read {file_path}")

    path_validation(file_path, check_for_file=True)

    # Handle JSON file read
    if file_path.suffix == ".json":
        with open(file_path, "r") as f:
            content = json.load(f)

    # Handle YAML file read
    elif file_path.suffix == ".yaml":
        with open(file_path, "r") as f:
            content = yaml.safe_load(f)

    # Handle geojson file read
    elif file_path.suffix == ".geojson":
        with open(file_path, "r") as f:
            content = geojson.load(f)

    else:
        logger.error(
            f"Could not read the {file_path}, this feature is not yet implemented"
        )
        raise FeatureNotImplementedError(
            f"File of type {file_path.suffix} \
            is not yet implemented for reading purpose"
        )

    logger.debug(f"{file_path} read successfully")
    return content


def write_file(content: dict, file_path: str, **kwargs) -> None:
    """Utility function to write to a file..

    Supports json, yaml and geojson.

    Args:
        content (dict): Python dict content
        file_path (str): Path to a file to be read
        kwargs (dict): Keyword arguments passed to
            relevant writer.

    Raises:
        FeatureNotImplementedError: Raises if invalid file type is passed.
    """
    file_path = Path(file_path)
    path_validation(file_path.parent)

    # Handle JSON file write
    if file_path.suffix == ".json":
        with open(file_path, "w") as f:
            json.dump(content, f, **kwargs)

    # Handle YAML file write
    elif file_path.suffix == ".yaml":
        with open(file_path, "w") as f:
            yaml.safe_dump(content, f, **kwargs)

    # Handle geojson file write
    elif file_path.suffix == ".geojson":
        with open(file_path, "w") as f:
            geojson.dump(content, f, **kwargs)

    else:
        raise FeatureNotImplementedError(
            f"File of type {file_path.suffix} \
            is not yet implemented for writing purpose"
        )




from click import format_filename
import matplotlib.pyplot as plt
from erad.scenarios.utilities import ProbabilityFunctionBuilder, GeoUtilities
from erad.constants import ELEVATION_RASTER_FILE, DATA_FOLDER, FLOOD_HISTORIC_SHP_PATH
from shapely.geometry import MultiPolygon, Point, LineString, Polygon
from erad.scenarios.abstract_scenario import BaseScenario
from datetime import datetime, timedelta
from scipy.optimize import curve_fit
from scipy.spatial import Delaunay
import geopandas as gpd
import pandas as pd
import numpy as np
import stateplane
import xmltodict
import itertools
import requests
import rasterio
import tarfile
import time
import sys
import os

from erad.scenarios.common import AssetTypes, asset_list
from erad.scenarios.utilities import ProbabilityFunctionBuilder

#plt.ion()
class FlooadScenario(BaseScenario, GeoUtilities):
    """Base class for FlooadScenario. Extends BaseScenario and GeoUtilities

    Attributes:
        origin (Point): Earthquake origin point
        probability_model (dict): Dictionary mapping asset types to probability funcitons
        timestamp (datetime): Scenario occurance time 
        kwargs (dict): Additional parameters relevant for a particular scenario type      
    """
    
    fragility_curves = {
        # Electrical Grid Risk Assessment Against Flooding in Barcelona and Bristol Cities
        AssetTypes.substation.name : ProbabilityFunctionBuilder("norm", [1.5, 0.2]),
        AssetTypes.solar_panels.name : ProbabilityFunctionBuilder("norm", [0.04, .01]),
        # AssetTypes.buried_lines.name : ProbabilityFunctionBuilder("norm", [80, 10]),
        # estimated
        AssetTypes.wind_turbines.name : ProbabilityFunctionBuilder("lognorm", [0.8, 3, 2]),
        AssetTypes.battery_storage.name : ProbabilityFunctionBuilder("norm", [0.04, .01]),
        # estimated
        AssetTypes.transmission_poles.name  :  ProbabilityFunctionBuilder("lognorm", [0.8, 5, 3]),
        # Tsunami Fragility Functions for Road and Utility Pole Assets Using Field Survey and Remotely Sensed Data from the 2018 Sulawesi Tsunami, Palu, Indonesia
        AssetTypes.distribution_poles.name  : ProbabilityFunctionBuilder("lognorm", [1, 0.2, 1]),
        AssetTypes.transmission_overhead_lines.name  : ProbabilityFunctionBuilder("norm", [0.04, .01]),    
        AssetTypes.distribution_overhead_lines.name  : ProbabilityFunctionBuilder("norm", [0.04, .01]),
    }
    
    
    def __init__(self,  poly : MultiPolygon , probability_model : dict, timestamp : datetime, **kwargs) -> None:
        super(FlooadScenario, self).__init__(poly, probability_model, timestamp, **kwargs)
        self.kwargs = kwargs
        self.samples = 100
        self.use_api = True
        
        # self.map_elevation() 
        if 'type' in kwargs and kwargs['type'] == 'live':
            self.flows = pd.DataFrame()
            self.levels = pd.DataFrame()
            self.real_time()
        else:
            from shapely import wkt
            self.flows = pd.read_csv(kwargs['file_flow'],index_col=0, parse_dates=True)
            self.levels = pd.read_csv(kwargs['file_levels'], index_col=0, parse_dates=True)
            df = pd.read_csv(kwargs['file_gaugues'])
            df['geometry'] = df['geometry'].apply(wkt.loads)
            crs = {'init': 'epsg:4326'}
            self.gauges = gpd.GeoDataFrame(df).set_geometry('geometry')
            #self.gauges = gpd.read_file(kwargs['file_gaugues'])
            pass
        # self.plot = DynamicUpdate(self.X, self.Y, self.Z, self.levels)
        return
    
    @classmethod
    def from_historical_flood_by_code(cls, flood_code : str, probability_function : dict = None):
        data_file = os.path.join(DATA_FOLDER, FLOOD_HISTORIC_SHP_PATH)
        assert os.path.exists(data_file), f"The data file {data_file} not found"
        flood_data = gpd.read_file(data_file)
        flood_data = flood_data[flood_data['DFIRM_ID'] == flood_code]
        raise NotImplementedError("Model has not been implemented")
        
    def real_time(self):
        self.gauges = self.get_flow_measurements(0)
        self.gauges.to_csv("gauges.csv")
        flows, levels = self.gauges_in_polygon()
        for c in flows.columns:
            flows[c] = flows[c].interpolate(method='polynomial', order=2)
            levels[c] = levels[c].interpolate(method='polynomial', order=2)
        flows = flows.ffill(axis = 0)
        levels = levels.ffill(axis = 0)
        flows = flows.bfill(axis = 0)
        levels = levels.bfill(axis = 0)
        self.flows = flows.resample("15T").interpolate()
        self.levels = levels.resample("15T").interpolate()
        self.flows.to_csv("flows.csv")
        self.levels.to_csv("levels.csv")

    @property
    def valid_timepoints(self):
        return list(self.flows.index)
    
    def gauges_in_polygon(self):
        all_flows = pd.DataFrame()
        all_levels = pd.DataFrame()
        for idx, gauge in self.gauges.iterrows():
            flow = pd.DataFrame()
            level = pd.DataFrame()
            gauge_id = gauge['GaugeLID']
            url = f'https://water.weather.gov/ahps2/hydrograph_to_xml.php?gage={gauge_id}&output=xml'
            response = requests.get(url)
            #print(response.text)
            data = xmltodict.parse(response.text)
            times_series_data = data['site']['observed']['datum']
            time_points = []
            flow_points = []
            level_points =[]
            for data_point in times_series_data:
                time_points.append(data_point['valid']['#text'])
                flow_points.append(float(data_point['secondary']['#text']))
                level_points.append(float(data_point['primary']['#text']))
            index =pd.to_datetime(time_points)
            flow.index =index
            flow[gauge_id] = flow_points
            level.index =index
            level[gauge_id] = level_points
            print(flow)
            print(level)
            all_flows = flow.merge(all_flows, right_index=True, left_index=True, how="outer")
            all_levels = level.merge(all_levels, right_index=True, left_index=True, how="outer")
        return all_flows, all_levels
    
    def get_flow_measurements(self, forecast_day: int = 0):
        if forecast_day == 0:
            forecast_tag = "obs"
        else:
            if forecast_day == 1:   
                forecast_tag = "fcst_f024"
            else:
                forecast_tag = f"fcst_f{forecast_day*24}"
                
        url = f"https://water.weather.gov/ahps/download.php?data=tgz_{forecast_tag}"
        r = requests.get(url, allow_redirects=True)
        date = str(datetime.now()).replace(":", "_").replace(".", "_").replace(" ", "_")
        shape_file_path = os.path.join(DATA_FOLDER, f'flood_shapefile_{date}.tgz')
        open(shape_file_path, 'wb').write(r.content)
        file_path = self.extract(shape_file_path, forecast_tag)
        data = gpd.read_file(file_path)
        if not data.empty:
            water_info = []
            for idx, row in data.iterrows():
                point = Point(row['Latitude'], row['Longitude'])
                if self.multipolygon.contains(point):
                    water_info.append(row)
            return gpd.GeoDataFrame(water_info)
        else:
            raise Exception("No water measurement found in selected area.")
    
    def extract(self, tar_url, ext):
        extract_path = os.path.join(DATA_FOLDER, "flood_shape_file")
        tar = tarfile.open(tar_url, 'r')
        for item in tar:
            tar.extract(item, extract_path)
            if item.name.find(".tgz") != -1 or item.name.find(".tar") != -1:
                self.extract(item.name, "./" + item.name[:item.name.rfind('/')])
        return os.path.join(extract_path, f'national_shapefile_{ext}.shp')
    
    @classmethod
    def from_live_data(
        cls,
        poly : MultiPolygon, 
        probability_function : dict, 
        startrime: datetime, 
        duration: timedelta, 
        timestep: timedelta
        ):
        kwargs = {
            'startrime' : startrime, 
            'duration' : duration,
            'timestep' : timestep,
            'type' : 'live',
        }
        
        return cls(poly, probability_function, startrime, **kwargs)
   
    def get_elevation_by_latlong(self, lat, lon):
        coords = ((lat,lon), (lat,lon))
        vals = self.raster.sample(coords)
        for val in vals:
            elevation=val[0]
            return 255-elevation

    def map_elevation(self):
        y_min, x_min, y_max, x_max = self.multipolygon.bounds
        ys = np.linspace(y_min, y_max, self.samples, endpoint=True)
        xs = np.linspace(x_min, x_max, self.samples, endpoint=True)
        self.X, self.Y = np.meshgrid(xs, ys)
        
        if self.use_api:
            self.Z = self.create_elevation_profile_using_api()
        else:
            self.Z = self.create_elevation_profile_using_rasterfile() 
  
        r, c = self.X.shape
        for i in range(r):
            for j in range(c):
                x_sp, y_sp = stateplane.from_lonlat(self.X[i, j], self.Y[i, j])
                self.X[i, j] =x_sp
                self.Y[i, j] =y_sp
        
        self.min_elevation = np.min(self.Z)
        self.max_elevation = np.max(self.Z)       

        pts = np.array([self.X.flatten(), self.Y.flatten(), self.Z.flatten()]).T
        self.volume = self.calc_polyhedron_volume(pts)

        return self.X, self.Y, self.Z

    def create_elevation_profile_using_rasterfile(self):
        raster_file = os.path.join(DATA_FOLDER, ELEVATION_RASTER_FILE)
        self.raster = rasterio.open(raster_file)
        r, c = self.X.shape
        self.Z = np.zeros(self.X.shape)
        for i in range(r):
            for j in range(c):
                self.Z[i, j] = self.get_elevation_by_latlong(self.X[i, j], self.Y[i, j])
        return self.Z
    
    def create_elevation_using_api(self, lat, lon):
        response = requests.get(
            f"https://api.airmap.com/elevation/v1/ele?points={lat},{lon}"
            )
        response = response.json()
        if response['status'] == 'success':
            # print(response, '--')
            return response['data'][0] * 3.28084
            
        else:
            print(response)
        
    def create_elevation_profile_using_api(self, X=None, Y=None):
        coords = []
        inputs_passed = True
        if X is None or Y is None:
            inputs_passed = False
            X = self.X.flatten()
            Y = self.Y.flatten()
        
        for x, y in zip(X, Y):
            coords.extend([str(y), str(x)])
 
        offset = 1000
        Z_coords = []
        for i in range(int(len(coords)/offset) + 1):
         
            filt_coords = coords[i * offset: (i + 1) * offset]
            filt_coords = ",".join(filt_coords)    
            response = requests.get(
            f"https://api.airmap.com/elevation/v1/ele?points={filt_coords}"
            )
            response = response.json()
            if response['status'] == 'success':
                Z_coords.extend(response['data'])
                # print(response)
            else:
                print(response)
        
        Z = np.array(Z_coords)
        if inputs_passed:
            return Z
        else:
            Z = np.reshape(Z, self.X.shape) * 3.28084
            return (Z)

    def calc_polyhedron_volume(self, pts):

        def tetrahedron_volume(a, b, c, d):
            return np.abs(np.einsum('ij,ij->i', a-d, np.cross(b-d, c-d))) / 6
        
        dt = Delaunay(pts)
        tets = dt.points[dt.simplices]
        self.polyhedron_volume = np.sum(tetrahedron_volume(tets[:, 0], tets[:, 1], 
                                        tets[:, 2], tets[:, 3]))
        
        return self.polyhedron_volume
   
    @property
    def area(self) -> float:
        """Method to calculate area of affected region."""
        raise NotImplementedError("Method needs to be defined in derived classes")
    
    @property
    def polygon(self) -> MultiPolygon:
        """Method to return polygon for the affected region."""
        raise NotImplementedError("Method needs to be defined in derived classes")
        
    @property
    def boundary(self) -> LineString:
        """Method to return boundary for the affected region."""
        raise NotImplementedError("Method needs to be defined in derived classes")
        
    @property
    def centroid(self) -> Point:
        """Method to return the centroid of the affected region."""
        return self.multipolygon.centroid
    
    def increment_time(self) -> dict:
        """Method to increment simulation time for time evolviong scenarios."""
        raise NotImplementedError("Method needs to be defined in derived classes")

    def calculate_survival_probability(self, assets : dict, timestamp: datetime) -> dict:
        """Method to calculate survival probaility of asset types.

        Args:
            assets (dict): The dictionary of all assets and their corresponding asset types
        """
        print('Calculating survival probaiblity ...')
        water_elevations = []
        coords = [
            [],[],[]
        ]
        z = []
        for idx, row in self.gauges.iterrows():
            
            gauge = row['GaugeLID']
            level = self.levels[gauge][timestamp] 
            
            lat = row['Latitude']
            lon = row['Longitude'] 
            x_i, y_i = stateplane.from_lonlat(lon, lat)
            coords[0].append(x_i)
            coords[1].append(y_i)
            
            
            if self.use_api:
                # elevation = self.create_elevation_using_api(lat, lon)
                # TODO: Don't know the different between level and elevation or
                # how they relate 
                elevation = row['Elevation_ft']
                water_elevation = float(level) + elevation
            else:
                elevation = self.get_elevation_by_latlong(lat, lon)
                water_elevation = self.get_elevation_by_latlong(lat, lon) + float(level)

            self.gauges["elevation"] = elevation
            z.append(water_elevation) #water_elevation, flow
            coords[2].append(f"Gauge: {gauge}\nElevation: {elevation}\nWater level: {level}")
            water_elevations.append(water_elevation)
        self.gauges["water_level"] = water_elevations
        
        m = self.polyfit2d(np.array(coords[0]), np.array(coords[1]), np.array(z))
        # self.z_ = np.zeros(self.X.shape)
        
        for asset_type, asset_dict in assets.items():
            Xs = []
            Ys = []
            for asset, asset_data in asset_dict.items():
                Xs.append(asset_data['coordinates'][1])
                Ys.append(asset_data['coordinates'][0])
                x_i, y_i = stateplane.from_lonlat(asset_data['coordinates'][1], asset_data['coordinates'][0])
                z_i = self.polyval2d(x_i, y_i, m)
                assets[asset_type][asset]['asset_water_level_ft'] = z_i.tolist()  #list(self.polyval2d(x_i, y_i, m))
            
            # if self.use_api:    
            #     asset_elevations_ft = self.create_elevation_profile_using_api(Xs, Ys)
            # else:
            #     asset_elevations_ft = self.get_elevation_by_latlong(lat, lon)

            i = 0
            for asset, asset_data in asset_dict.items():
                h =   asset_data['asset_water_level_ft'] - asset_data['elevation_ft']
                assets[asset_type][asset]['submerge_depth_ft'] = h
                if asset_type in self.probability_model:
                    probability_function = self.probability_model[asset_type]
                    failure_probability = probability_function.probability(h)
                    assets[asset_type][asset]["survival_probability"] = 1 - failure_probability
                else:
                    assets[asset_type][asset]["survival_probability"] = 1
                i+=1
        
        return assets

    def function_to_fit(self, data, a, b, c):
        x = data[0]
        y = data[1]
        return (a * x) + (y * b) + c
    
    def polyfit2d(self, x, y, z, order=3):
        ncols = (order + 1)**2
        G = np.zeros((x.size, ncols))
        ij = itertools.product(range(order+1), range(order+1))
        for k, (i,j) in enumerate(ij):
            G[:,k] = x**i * y**j
        m, _, _, _ = np.linalg.lstsq(G, z)
        return m

    def polyval2d(self, x, y, m):
        order = int(np.sqrt(len(m))) - 1
        ij = itertools.product(range(order+1), range(order+1))
        z = np.zeros_like(x)
        for a, (i,j) in zip(m, ij):
            z += a * x**i * y**j
        return z


class DynamicUpdate():
    #Suppose we know the x range
    min_x = 0
    max_x = 10

    def __init__(self, X, Y, Z, flows):
        self.X = X
        self.Y = Y
        self.Z = Z
        self.flows = flows
        self.min_elevation = np.min(Z)
        self.max_elevation = np.max(Z) 
        
        ncontours = 15
        step_size = (self.max_elevation - self.min_elevation) / ncontours
        self.levels = np.arange(self.min_elevation, self.max_elevation, step_size)
        self.fig = plt.figure()
        self.ax1 = self.fig.add_subplot(121, projection='3d')        
        self.ax2 = self.fig.add_subplot(222)
        self.ax3 = self.fig.add_subplot(224)

        self.ax1.plot_surface(X, Y, Z, rstride=1, cstride=1, color='0.99', antialiased=True, edgecolor='0.5')

        self.ax2.contour(X, Y, Z, zdir='z', cmap='coolwarm', levels=self.levels)
        self.flows.plot(ax = self.ax3)
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()
        plt.show()
        self.fig.savefig(f"topology_0.png")


    def update(self, water_elevations, scatter_points, water_surface, timestamp):
        self.ax1.clear()
        self.ax2.clear()
        self.ax3.clear()
        
        X = self.X
        Y = self.Y
        Z = self.Z
        
        w = np.where(water_surface > Z, water_surface, np.nan)
        w.flat[0] = np.nan
        self.ax1.plot_surface(X, Y, w, rstride=1, cstride=1, color='c', antialiased=True, alpha=0.5, shade=False)
        self.ax1.plot_surface(X, Y, Z, rstride=1, cstride=1, color='0.99', antialiased=True, edgecolor='0.5')
        
        self.ax2.contour(X, Y, Z, zdir='z', cmap='coolwarm', levels=self.levels)
        self.ax2.scatter(scatter_points[0], scatter_points[1], color='red')
        
        for x, y, t in zip(scatter_points[0], scatter_points[1], scatter_points[2]):
            self.ax2.text(x, y, t, fontsize=8)
            print(t)
        
        self.flows.plot(ax = self.ax3)
        self.ax3.axvline(timestamp, color="r")
        
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()
        #self.fig.savefig(f"topology_{int(water_elevation)}.png")
       

if __name__ == '__main__':

    
    
    from erad.scenarios.common import asset_list
    
    assets, multiploygon = asset_list(38.46, -122.95, 38.53, -122.80)  
    flood_1 = FlooadScenario(
        multiploygon, 
        None, 
        None, 
        file_flow=r'C:\Users\alatif\Documents\GitHub\erad\erad\scenarios\flows.csv',
        file_levels=r'C:\Users\alatif\Documents\GitHub\erad\erad\scenarios\levels.csv',
        file_gaugues=r'C:\Users\alatif\Documents\GitHub\erad\erad\scenarios\gauges.csv',
        )
    timestamp = flood_1.valid_timepoints[0]
    assets = flood_1.calculate_survival_probability(assets, timestamp)
    print(assets)
    # flood_1 = FlooadScenario.from_live_data(multiploygon, None, None, None, None)
    # timestamp = flood_1.valid_timepoints[-1]
    # for timestamp in flood_1.valid_timepoints:
    #     assets = flood_1.calculate_survival_probability(assets, timestamp)
    #     print(assets)
    #     time.sleep(0.01)
    


""" Module for storing pydantic model for injecting scenarios. """

from typing import Dict, Optional, List
import datetime

from pydantic import BaseModel, validator, confloat


class PointEarthquake(BaseModel):
    longitude: confloat(ge=-180, le=180)
    latitude: confloat(ge=-90, le=90)
    probability_model: Optional[Dict]
    timestamp: datetime.datetime
    magnitude: confloat(ge=0, le=10)
    depth: confloat(ge=0)

class PolygonFlooding(BaseModel):
    polygon: List
    probability_model: Optional[Dict]
    timestamp: datetime.datetime
    file_flow: str 
    file_levels: str 
    file_gaugues: str



""" This module contains functions and utilities to check for the
possibility of microgrid formation.
"""

from typing import List, Dict
import math
import json

from neo4j import GraphDatabase
import networkx as nx
import matplotlib.pyplot as plt


def create_directed_graph(
    driver: GraphDatabase.driver,
):
    """Creates a directed graph representation of the power network.

    For now we have read all the relationships and nodes. We will need to
    filter this by feeder to avoid running into memory issues in future for
    larger graph.

    Args:
        driver (GraphDatabase.driver): Instance of `GraphDatabase.driver`
            instance
    """

    # Get the buses, customers, pvs, energystorage and line sections
    power_network_query = """
    MATCH (sourceNode:Bus)-[relationship:CONNECTS_TO]-(targetNode:Bus)
    return relationship{.*} , sourceNode {.*}, targetNode{.*}
    """

    # Gettings relations between customers and buses
    customer_bus_network_query = """
    MATCH (sourceNode:Bus)-[relationship:CONSUMES_POWER_FROM]-(targetNode:Load)
    return relationship{.*} , sourceNode {.*}, targetNode{.*}
    """

    # Gettings relations between critical infrastructures and buses
    critical_infra_bus_network_query = """
    MATCH (sourceNode:Bus)-[relationship:GETS_POWER_FROM]-(targetNode)
    return relationship{.*} , sourceNode {.*}, targetNode{.*}
    """

    # Gettting relations between PVs and buses
    pv_bus_network_query = """
    MATCH (sourceNode:Bus)-[relationship:INJECTS_ACTIVE_POWER_TO]-(targetNode:Solar)
    return relationship{.*} , sourceNode {.*}, targetNode{.*}
    """

    # Getting relations between energy storage and buses
    es_bus_network_query = """
    MATCH (sourceNode:Bus)-[relationship:INJECTS_POWER]-(targetNode:EnergyStorage)
    return relationship{.*} , sourceNode {.*}, targetNode{.*}
    """

    relations = []
    for query in [
        power_network_query,
        customer_bus_network_query,
        pv_bus_network_query,
        es_bus_network_query,
        critical_infra_bus_network_query
    ]:

        with driver.session() as session:
            result = session.read_transaction(lambda tx: tx.run(query).data())
        relations.extend(result)

    graph = nx.Graph()
    for rel in relations:

        # Unpack the relationship data
        relationship = rel["relationship"]
        source_node = rel["sourceNode"]
        target_node = rel["targetNode"]

        # Add nodes if not already present in the graph
        for node in [source_node, target_node]:
            if not graph.has_node(node["name"]):
                graph.add_node(node["name"], **node)

        # Add relationship
        graph.add_edge(source_node["name"], target_node["name"], **relationship)

    return graph.to_directed()

def node_connected_to_substation(
    substation_nodes: List[str],
    driver: GraphDatabase.driver
):
    """ Gives list of nodes still connected to substation. """
    directed_graph = create_directed_graph(driver)
    edges_to_be_removed = []

    for edge in directed_graph.edges():
        edge_data = directed_graph.get_edge_data(*edge)
        if "survive" in edge_data and int(edge_data["survive"]) == 0:
            edges_to_be_removed.append(edge)

    if edges_to_be_removed:
        directed_graph.remove_edges_from(edges_to_be_removed)
        wcc = nx.weakly_connected_components(directed_graph)

        for _, weak_component in enumerate(wcc):
            wcc_graph = directed_graph.subgraph(weak_component)
            nodes = wcc_graph.nodes()
            for sub_node in substation_nodes:
                if sub_node in nodes:
                    return nodes
    else:
        nodes = []
        for edge in directed_graph.edges():
            nodes.extend(edge)
        return nodes
    return []
            

def check_for_microgrid(driver: GraphDatabase.driver, output_json_path: str):
    """Checks for possibility of microgrid in each subgraph.

    Args:
        driver (GraphDatabase.driver): Instance of `GraphDatabase.driver`
            instance
       output_json_path (str): JSON file path for exporting the metric.
    """

    directed_graph = create_directed_graph(driver)
    node_data = {item[0]: item[1] for item in directed_graph.nodes(data=True)}

    edges_to_be_removed = []
    subgraphs = {}

    for edge in directed_graph.edges():
        edge_data = directed_graph.get_edge_data(*edge)
        if "survive" in edge_data and int(edge_data["survive"]) == 0:
            edges_to_be_removed.append(edge)

    if edges_to_be_removed:
        directed_graph.remove_edges_from(edges_to_be_removed)
        wcc = nx.weakly_connected_components(directed_graph)

        for id, weak_component in enumerate(wcc):

            # Let's create a networkx representation to perform
            # multiple source multiple sink max flow problem
            # https://faculty.math.illinois.edu/~mlavrov/docs/482-fall-2019/lecture27.pdf
            source_capacity, sink_capacity = 0, 0
            wcc_graph = directed_graph.subgraph(weak_component)
            wcc_graph = nx.DiGraph(wcc_graph)

            for new_node in ["infinity_source", "infinity_sink"]:
                if not wcc_graph.has_node(new_node):
                    wcc_graph.add_node(new_node)

            sinks, sources = [], []
            for node in wcc_graph.nodes():
                # Connect all loads to infinity sink

                if "pv" in node or "es_" in node or node_data.get(node, {}).get('backup', None) == 1:
                    wcc_graph.add_edge(node, "infinity_source", capacity=1e9)
                    wcc_graph.add_edge("infinity_source", node, capacity=1e9)
                    sources.append(node)

                    cap_ = None 
                    if 'kw' in node_data[node]:
                        cap_ = node_data[node]["kw"]
                    elif 'capacity' in node_data[node]:
                        cap_ = node_data[node]["capacity"]
                    elif 'backup_capacity_kw' in node_data[node]:
                        cap_ = node_data[node]["backup_capacity_kw"]
                    else:
                        raise Exception('Not a valid source!')
                    source_capacity += cap_

                elif "load" in node or node_data.get(node, {}).get('survive', None) is not None :
                    wcc_graph.add_edge(node, "infinity_sink", capacity=1e9)
                    wcc_graph.add_edge("infinity_sink", node, capacity=1e9)
                    sinks.append(node)
                    sink_capacity += math.sqrt(
                    node_data[node].get('kW', 0) ** 2
                    + node_data[node].get('kvar', 0) ** 2
                    ) * float(node_data[node].get("critical_load_factor", 0))
                   
            
            # if id == 2:
            #     breakpoint()
            flow_value, _ = nx.maximum_flow(
                wcc_graph, "infinity_source", "infinity_sink", capacity="kva"
            )
    
            subgraphs[f"weak_component_{id}"] = {
                "length": len(weak_component),
                "max_flow": flow_value,
                "sources": sources,
                "sinks": sinks,
                "source_capacity": source_capacity,
                "sink_capacity": sink_capacity,
            }



    if output_json_path:
        with open(output_json_path, "w") as fpointer:
            json.dump(subgraphs, fpointer)

    return subgraphs



""" Module for parsing Homeland infrastructure foundation level-data.

Idea is to take the bounding box and find the subset of 
infrastructure in that region.
"""
# standard imports
from pathlib import Path
import math
from typing import Union, List

# third-party imports
import pandas as pd
import stateplane

# internal imports
from erad.utils.util import path_validation


def get_subset_of_hifld_data(
    csv_file: str,
    bounds: List,
    output_folder: str,
    logitude_column_name: str = "X",
    latitude_column_name: str = "Y",
    columns_to_keep: List[str] = ["X", "Y"],
    name_of_csv_file: Union[str, None] = None,
) -> None:
    """Extracts a subset of HIFLD data set.

    Args:
        csv_file (str): Path to HIFLD data csv file
        bounds (List): Bounding box coordinates
        output_folder (str): Path to output folder
        logitude_column_name (str): Expects column with name 'X'
        latitude_column_name (str): Expects column with name 'Y'
        columns_to_keep (List): List of column names to keep
            by default keeps all of them
       name_of_csv_file (Union[str, None]): Name of csv file to export
            filtered set
    """

    # Unpacking the bounds data
    longitude_min, latitude_min, longitude_max, latitude_max = bounds

    # Do a path validation
    csv_file = Path(csv_file)
    output_folder = Path(output_folder)
    path_validation(csv_file, check_for_file=True, check_for_file_type=".csv")
    path_validation(output_folder)

    # Reading the hifld csv data
    df = pd.read_csv(csv_file)

    # filtering for bounds
    df_filtered = df[
        (df[logitude_column_name] >= longitude_min)
        & (df[logitude_column_name] <= longitude_max)
        & (df[latitude_column_name] >= latitude_min)
        & (df[latitude_column_name] <= latitude_max)
    ]

    # Keep only the limited columns
    df_subset = df_filtered[columns_to_keep]

    # export the subset
    file_name = name_of_csv_file if name_of_csv_file else csv_file.name
    df_subset.to_csv(output_folder / file_name)


def get_relationship_between_hifld_infrastructures(
    hifld_data_csv: str,
    unique_id_column: str,
    load_csv: str,
    bus_csv: str,
    output_csv_path: str,
    distance_threshold: float = 2000.0,
):
    """Creates a relationship between consumers and HIFLD infrastructures.

    Args:
        hifld_data_csv (str): Path to filtered HIFLD data csv file
        unique_id_column (List): Column name used as identifier
            for critical infrastructures
        load_csv (str): Path to load csv file
        bus_csv (str): Path to bus csv file
        output_csv_path (str): output csv path for storing relationship csv
        distance_threshold (float): Distance threshold used for mapping
            customer to critical infrastructure
    """
    hifld_data_csv = Path(hifld_data_csv)
    bus_csv = Path(bus_csv)
    load_csv = Path(load_csv)
    output_csv_path = Path(output_csv_path)

    path_validation(
        hifld_data_csv, check_for_file=True, check_for_file_type=".csv"
    )
    path_validation(bus_csv, check_for_file=True, check_for_file_type=".csv")
    path_validation(load_csv, check_for_file=True, check_for_file_type=".csv")
    path_validation(output_csv_path.parents[0])

    hifld_data_df = pd.read_csv(hifld_data_csv)
    load_df = pd.read_csv(load_csv)
    bus_df = pd.read_csv(bus_csv)

    merged_data = pd.merge(
        load_df, bus_df, how="left", left_on="source", right_on="name"
    ).to_dict(orient="records")

    # Container for storing shelter relationships
    _relationship = []
    for _record in hifld_data_df.to_dict(orient="records"):
        _lon, _lat = _record["LONGITUDE"], _record["LATITUDE"]

        # convert into state plane coordinates
        _lon_translated, _lat_translated = stateplane.from_lonlat(_lon, _lat)

        # Loop through all the loads
        for load_record in merged_data:

            load_lon, load_lat = (
                load_record["longitude"],
                load_record["latitude"],
            )

            # convert into state plane coordinates
            load_lon_translated, load_lat_translated = stateplane.from_lonlat(
                load_lon, load_lat
            )

            # computes distance
            distance = math.sqrt(
                (_lat_translated - load_lat_translated) ** 2
                + (_lon_translated - load_lon_translated) ** 2
            )

            if distance < distance_threshold:
                _relationship.append(
                    {
                        unique_id_column: _record[unique_id_column],
                        "load_name": load_record["name_x"],
                        "distance": distance,
                    }
                )

    df = pd.DataFrame(_relationship)
    df.to_csv(output_csv_path)



"""Module contains class and utility functions to manage
interactions with Neo4J database. 
"""

# standard libraries
import os
import logging
from typing import Union


# third-party libraries
from dotenv import load_dotenv
from neo4j import GraphDatabase, basic_auth

# internal imports
from erad.db.credential_model import Neo4jConnectionModel


load_dotenv()
NEO4J_URL = os.getenv("NEO4J_URL")
NEO4J_USERNAME = os.getenv("NEO4J_USERNAME")
NEO4J_PASSWORD = os.getenv("NEO4J_PASSWORD")

logger = logging.getLogger(__name__)


class Neo4J:
    """Class for managing interaction with Neo4J database.

    Attributes:
        neo4j_url (str): URL for connecting to Neo4j
        neo4j_username (str): Username for Neo4j database
        neo4j_password (str): Password for Neo4j database
        use_env (bool): True if above info are to be collected
            from env file.
        driver (GraphDatabase.driver): Neo4J driver instance
    """

    def __init__(
        self,
        neo4j_url: Union[str, None] = None,
        neo4j_username: Union[str, None] = None,
        neo4j_password: Union[str, None] = None,
    ) -> None:
        """Constructor for Neo4J class.

        Args:
            neo4j_url (str): URL for connecting to Neo4j
            neo4j_username (str): Username for Neo4j database
            neo4j_password (str): Password for Neo4j database
        """

        self.neo4j_url = NEO4J_URL if NEO4J_URL else neo4j_url
        self.neo4j_username = (
            NEO4J_USERNAME if NEO4J_USERNAME else neo4j_username
        )
        self.neo4j_password = (
            NEO4J_PASSWORD if NEO4J_PASSWORD else neo4j_password
        )

        connection = Neo4jConnectionModel(
            neo4j_url=self.neo4j_url,
            neo4j_username=self.neo4j_username,
            neo4j_password=self.neo4j_password,
        )

        self.driver = GraphDatabase.driver(
            connection.neo4j_url,
            auth=basic_auth(
                connection.neo4j_username, connection.neo4j_password
            ),
        )

        logger.debug(
            f"Connected to {connection.neo4j_url} database successfully"
        )

    @staticmethod
    def rename_labels(label):
        """Method to replace the invalid character."""
        invalid_chars = ["-", ":", "(", ")", "."]
        for invalid_char in invalid_chars:
            if invalid_char in label:
                label = label.replace(invalid_char, "__")
        return label

    # def add_node(
    #     self,
    #     labels: Union[List, None] = None,
    #     properties: Union[Dict, None] = None,
    # ) -> None:
    #     """Method to add node to the Neo4j database.

    #     Args:
    #         labels (Union[List, None]): List of labels to be used for node
    #         properties (Union[Dict, None]): Properties to be used for the node
    #     """
    #     if labels is None:
    #         labels = []

    #     if properties is None:
    #         properties = {}

    #     labels = ":".join([self.rename_labels(label) for label in labels])
    #     cypher_query = "CREATE (:" + labels + " $properties)"

    #     with self.driver.session() as session:
    #         session.write_transaction(
    #             lambda tx: tx.run(cypher_query, properties=properties)
    #         )

    # def add_relationship(
    #     self,
    #     from_node_label: str,
    #     to_node_label: str,
    #     from_node: str,
    #     to_node: str,
    #     relationship_label: str,
    #     relationship_properties: Union[Dict, None] = None,
    # ) -> None:
    #     """Method to create relationship in graph database.

    #     Args:
    #         from_node_label (str): Node label for from node
    #         to_node_label (str): Node label for to node
    #         from_node (str): From node name
    #         to_node (str): To node name
    #         relationship_label (str): Relationship label name
    #         relationship_properties (Union[Dict, None]): Properties to be used
    #             for relationship
    #     """

    #     if relationship_properties is None:
    #         relationship_properties = {}

    #     cypher_query = (
    #         f"MATCH (a:{from_node_label}),(b:{to_node_label}) WHERE a.name = '{from_node}'"
    #         + f" AND b.name = '{to_node}' CREATE (a)-[:"
    #         + self.rename_labels(relationship_label)
    #         + " $properties]->(b)"
    #     )

    #     with self.driver.session() as session:
    #         session.write_transaction(
    #             lambda tx: tx.run(
    #                 cypher_query, properties=relationship_properties
    #             )
    #         )

    # def read_query(self, cypher_query: str) -> List:
    #     """Executes a Cypher read query."""

    #     with self.driver.session() as session:
    #         result = session.read_transaction(
    #             lambda tx: tx.run(cypher_query).data()
    #         )

    #     return result

    def close_driver(self):
        """Method to close the driver."""
        self.driver.close()



""" Module for handling graph plots. """

# standard imports
import os
import abc
from typing import List, Dict

# third-party libraries
import networkx as nx
import plotly.graph_objects as go
from dotenv import load_dotenv

# internal libraries
from erad.utils.util import path_validation

load_dotenv()


class AbstractGraphPlot(abc.ABC):
    """Abstract interface for developing subclass to plot network graph."""

    @abc.abstractmethod
    def add_network_data(self, *args, **kwargs):
        """Abstract method for adding network data."""

    @abc.abstractmethod
    def prepare_plot(self, *args, **kwargs):
        """Abstract method for preparing and showing teh plot"""


class PloltyGraph(AbstractGraphPlot):
    """Class for managing graph plot using Plotly.

    Attributes:
        access_token (str): MapBox API token
        style (str): MapBox style
        zoom_level (int): Zoom level for the plot
        data (List): Stores the data to be fed to plotly for plotting
        scatter_data (Dict): Stores longitudes and latitudes of nodes
            from network
        fig (go.Figure): Plotly graph objects figure instance
    """

    def __init__(
        self,
        access_token: str = None,
        style: str = "carto-darkmatter",
        zoom_level: int = 13,
    ) -> None:
        """Constructor for `PlotlyGraph` Subclass.

        Args:
            access_token (str): MapBox API token
            style (str): MapBox style
            zoom_level (int): Zoom level for the plot
        """

        if access_token:
            self.access_token = access_token
        else:
            self.access_token = os.getenv("MAPBOX_API_KEY")
        self.style = style
        self.zoom_level = zoom_level

        self.data = []

    def _get_map_centre(self, longitudes: List[float], latitudes: List[float]):
        """Returns map center."""
        return {
            "lon": sum(longitudes) / len(longitudes),
            "lat": sum(latitudes) / len(latitudes),
        }

    def add_network_data(
        self,
        network: nx.Graph,
        latitude_property: str = "lat",
        longitude_property: str = "long",
        node_color: str = "blue",
        line_color: str = "red",
    ) -> None:
        """Method to add network data to plot data.

        Args:
            network (nx.Graph): Networkx graph instance
            latitude_property (str): Property name to be
                used as latitude
            longitude_property (str): Property name to be
                used as longitude
            node_color (str): Color name to be used to plot
                nodes
            line_color (str): Color name to be used to plot
                line segments
        """

        # Add nodes
        self.scatter_data = {"latitudes": [], "longitudes": []}

        for node in network.nodes.data():

            # Storing the lat lons in scatter data
            # container
            self.scatter_data["latitudes"].append(node[1][latitude_property])
            self.scatter_data["longitudes"].append(node[1][longitude_property])

        # Stroing the edge data in container
        line_data = {"latitudes": [], "longitudes": []}
        node_data = {node[0]: node[1] for node in network.nodes.data()}

        for edge in network.edges():
            line_data["latitudes"].extend(
                [
                    node_data[edge[0]][latitude_property],
                    node_data[edge[1]][latitude_property],
                    None,
                ]
            )

            line_data["longitudes"].extend(
                [
                    node_data[edge[0]][longitude_property],
                    node_data[edge[1]][longitude_property],
                    None,
                ]
            )

        # Adding plots to plotly graph object
        self.data.append(
            go.Scattermapbox(
                mode="markers",
                lon=self.scatter_data["longitudes"],
                lat=self.scatter_data["latitudes"],
                marker={"size": 5, "color": node_color},
            )
        )

        self.data.append(
            go.Scattermapbox(
                mode="markers+lines",
                lon=line_data["longitudes"],
                lat=line_data["latitudes"],
                marker={"size": 0},
                line={"color": line_color},
            )
        )

    def add_scatter_points(
        self,
        latitudes: List[float],
        longitudes: List[float],
        color: str = "yellow",
        size: int = 5,
    ) -> None:
        """Method for scatter points to plot data.

        Args:
            latitudes (List[float]): List of latitude points
            longitudes (List[float]): List of longitude points
            color (str): Color to be used for scatter points
            size (int): Size of scatter points
        """

        self.data.append(
            go.Scattermapbox(
                mode="markers",
                lon=longitudes,
                lat=latitudes,
                marker={"size": size, "color": color},
            )
        )

    def add_polygon(
        self,
        latitudes: List[float],
        longitudes: List[float],
        fill: str = "toself",
    ) -> None:
        """Method for adding polygon to the plot.

        Args:
            latitudes (List[float]): List of latitude points
            longitudes (List[float]): List of longitude points
            fill (str): Accepted fill value by plotly
        """
        self.data.append(
            go.Scattermapbox(
                lon=longitudes, lat=latitudes, fill=fill, mode="lines"
            )
        )

    def prepare_plot(self, show: bool = True):
        """Method to prepare and show the plot.

        Args:
            show (bool): True if want to see the plot.
        """
        self.fig = go.Figure(data=self.data)
        self.fig.update_layout(margin={"r": 0, "t": 0, "l": 0, "b": 0})
        self.fig.update_mapboxes(
            {
                "accesstoken": self.access_token,
                "style": self.style,
                "center": self._get_map_centre(
                    self.scatter_data["longitudes"],
                    self.scatter_data["latitudes"],
                ),
                "zoom": self.zoom_level,
            }
        )

        if show:
            self.fig.show()

    def html_export(self, html_file_path: str):
        """Method for exporting plot as HTML file."""
        path_validation(html_file_path)
        self.fig.write_html(html_file_path)




""" This module contains functions to update 
survival probability and survive property
based on disaster event. """

# manage python imports
from datetime import datetime 
import random

from neo4j import GraphDatabase

from erad.db.utils import _run_read_query
from erad.metrics.check_microgrid import node_connected_to_substation


def _update_critical_infra_based_on_grid_access_fast(
    critical_infras,
    driver: GraphDatabase.driver   
):
    """ A faster function to update survive attribute 
    based on whether critical infra has access to grid power."""
    # Let's get substation 
    cypher_query = f"""
        MATCH (n:Substation)
        WHERE n.survive is NULL OR n.survive<>0 
        RETURN n.name
    """
    substations = _run_read_query(driver, cypher_query)
    substations = [item["n.name"]  for item in substations]
    
    nodes = node_connected_to_substation(substations, driver)

    # Get all critical infra and check if they are in above nodes
    for cri_infra in critical_infras:
        cypher_query = f"""
                        MATCH (c:{cri_infra})
                        RETURN c.longitude, c.latitude, c.name, c.backup
                    """
        infras = _run_read_query(driver, cypher_query)
        
        for infra in infras:
            
            cypher_write_query = f"""
                MATCH (c:{cri_infra})
                WHERE c.name = $cname
                SET c.survive = $survive
                SET c.survival_probability = $s_prob
            """

            with driver.session() as session:
                session.write_transaction(
                    lambda tx: tx.run(
                        cypher_write_query,
                        cname=infra['c.name'],
                        s_prob= 1 if infra['c.name'] in nodes or int(infra['c.backup'])==1 else 0,
                        survive=1 if infra['c.name'] in nodes or int(infra['c.backup'])==1 else 0 
                    )
                )



def _update_critical_infra_based_on_grid_access(
    critical_infras,
    driver: GraphDatabase.driver 
):
    """ A function to update survive attribute 
    based on whether critical infra has access to grid power."""

    # Let's get substation 
    cypher_query = f"""
        MATCH (n:Substation)
        WHERE n.survive is NULL OR n.survive<>0 
        RETURN n.name
    """
    substations = _run_read_query(driver, cypher_query)
    substations = [item["n.name"]  for item in substations]

    for cri_infra in critical_infras:
        cypher_query = f"""
                        MATCH (c:{cri_infra})
                        RETURN c.longitude, c.latitude, c.name, c.backup
                    """
        infras = _run_read_query(driver, cypher_query)
        
        for infra in infras:
            connected = []
            for substation in substations:

                _name = "{name:" + f'"{infra["c.name"]}"' + "}"
                substation_name = "{name:" + f'"{substation}"' + "}"
                cypher_query = f"""
                    MATCH 
                        (g:{cri_infra} {_name})-[:GETS_POWER_FROM]-(b:Bus),
                        (s:Substation {substation_name}),
                        p=shortestPath((b)-[:CONNECTED_TO*]-(s))
                    WHERE all(r in relationships(p) WHERE r.survive is NULL OR r.survive<>0 )
                    RETURN length(p)
                """

                if not infra['c.backup']: 
                    path = _run_read_query(driver, cypher_query)
                    connected.append(1 if path else 0) 
                else:
                    connected.append(1)
                

            cypher_write_query = f"""
                MATCH (c:{cri_infra})
                WHERE c.name = $cname
                SET c.survive = $survive
                SET c.survival_probability = $s_prob
            """

            with driver.session() as session:
                session.write_transaction(
                    lambda tx: tx.run(
                        cypher_write_query,
                        cname=infra['c.name'],
                        s_prob= int(any(connected)),
                        survive=int(any(connected)) 
                    )
                )

def _update_critical_infra(
        scenario,
        critical_infras,
        driver: GraphDatabase.driver,
        timestamp: datetime 
):

    critical_infras_items = {}
    assets = {}

    for infra in critical_infras:
        cypher_query = f"""
                        MATCH (c:{infra})
                        RETURN c.longitude, c.latitude, c.name
                    """
        critical_infras_items[infra] = _run_read_query(driver, cypher_query)

    
    for c_infra, data in critical_infras_items.items():
        assets[c_infra] =  {
            item["c.name"]: {
                "coordinates": [item["c.longitude"], item["c.latitude"]]
            } for item in data
            if all([item["c.longitude"], item["c.latitude"]])
        }

    survival_prob = scenario.calculate_survival_probability(assets, datetime.now())
    
    # update the survival probability
    for infra in critical_infras:
        cypher_query = f"""
                        MATCH (c:{infra})
                        WHERE c.name = $cname
                        SET c.survive = $survive
                        SET c.survival_probability = $s_prob
                    """

        with driver.session() as session:
            for cname, cdict in survival_prob[infra].items():
                session.write_transaction(
                    lambda tx: tx.run(
                        cypher_query,
                        cname=cname,
                        s_prob=cdict["survival_probability"],
                        survive=int(
                            random.random() < cdict["survival_probability"]
                        ),
                    )
                )


""" This module contains a function to get the distribution 
line assets and update it's survival probability and survive
attribute based on event scenario. 
"""

# Manage python imports
from datetime import datetime
import random
import json
import random

random.seed(20)

from neo4j import GraphDatabase

from erad.db.utils import _run_read_query


def _create_assets(lines):
    """ Takes the list of lines and convert into
    asset dictionary. """

    return {
            line["r.name"]: {
                "coordinates": [line["r.latitude"], line["r.longitude"]],
                "heights_ft": float(line["r.height_m"])*3.28084,
                "elevation_ft": random.randint(50, 400)
            }
            for line in lines
            if all([line["r.longitude"], line["r.latitude"]])
        }

def _update_distribution_lines_survival(
    survival_probability,
    driver: GraphDatabase.driver   
):
    """ Takes survival probabilty and update survival 
    probability and survive property. """

  

    cypher_query = """
                    MATCH (b1:Bus)-[r:CONNECTS_TO {name: $rname}]-(b2:Bus)
                    SET r.survive = $survive
                    SET r.survival_probability = $s_prob
                """
    with driver.session() as session:
        for rname, rdict in survival_probability.items():
            session.write_transaction(
                lambda tx: tx.run(
                    cypher_query,
                    rname=rname,
                    s_prob=rdict.get("survival_probability", 1),
                    survive=int(random.random() < rdict.get("survival_probability", 1)),
                )
            )


def _update_distribution_overhead_lines(scenario, driver: GraphDatabase.driver, timestamp: datetime):
    """Get overhead lines and update the survival probability."""

    # Cypher query to get overhead line segments
    cypher_query = """ 
                    MATCH (b1:Bus)-[r:CONNECTS_TO]-(b2:Bus)
                    WHERE r.ampacity IS NOT NULL AND r.type = 'overhead'
                    RETURN r.longitude, r.latitude, r.name, r.type, r.height_m
                """
    overhead_lines = _run_read_query(driver, cypher_query)

    if overhead_lines:
        assets = {"distribution_overhead_lines": _create_assets(overhead_lines)}

        survival_prob = scenario.calculate_survival_probability(assets, timestamp)

        # with open('surv.json', "w") as fp:
        #     json.dump(survival_prob,fp)
        _update_distribution_lines_survival(
            survival_prob['distribution_overhead_lines'],
            driver
        )

def _update_distribution_underground_lines(scenario, driver: GraphDatabase.driver, timestamp: datetime):
    """Get overhead lines and update the survival probability."""

    # Cypher query to get overhead line segments
    cypher_query = """ 
                    MATCH (b1:Bus)-[r:CONNECTS_TO]-(b2:Bus)
                    WHERE r.ampacity IS NOT NULL AND r.type = 'underground'
                    RETURN r.longitude, r.latitude, r.name, r.type, r.height_m
                """
    underground_lines = _run_read_query(driver, cypher_query)

    if underground_lines:
        assets = {"buried_lines": _create_assets(underground_lines)}

        survival_prob = scenario.calculate_survival_probability(assets, timestamp)
        _update_distribution_lines_survival(
            survival_prob['buried_lines'],
            driver
        )




from shapely.geometry import MultiPolygon, Point,Polygon
from random import random,seed
from enum import IntEnum
import numpy as np

class ScenarioTypes(IntEnum):
    flood_m = 0
    wind_m_per_s = 1
    fire_m = 2
    earthquake_pga = 3
    
class AssetTypes(IntEnum):
    substation = 0
    solar_panels = 1
    buried_lines = 2
    wind_turbines= 3
    battery_storage = 4
    transmission_poles = 5
    distribution_poles = 6
    transmission_overhead_lines = 7
    distribution_overhead_lines = 8
    #substructures
    #conduit_burial

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_
    
    @classmethod
    def has_asset(cls, asset):
        print(asset)
        return asset in cls.__members__


def asset_list(x1=41.255, y1=-117.33, x2=41.255, y2=-117.33, samples=100):

    x = np.linspace(x1, x2, samples)
    y = np.linspace(y1, y2, samples)
    
    seed(3)
    asset_probabilities = {
        AssetTypes.substation: 1 / 10000.0,
        AssetTypes.solar_panels : 1/500,
        AssetTypes.buried_lines : 1/10.0,
        AssetTypes.wind_turbines : 1/5000,
        AssetTypes.battery_storage :1/2000,
        AssetTypes.transmission_poles: 1 / 10.0,
        AssetTypes.distribution_poles : 1 / 10.0,
        AssetTypes.transmission_overhead_lines : 1/10.0,
        AssetTypes.distribution_overhead_lines : 1/10.0,
    }
    
    heights_ft = {
        AssetTypes.substation.name : 3,
        AssetTypes.solar_panels.name : 10,
        AssetTypes.buried_lines.name : -3,
        AssetTypes.wind_turbines.name : 25,
        AssetTypes.battery_storage.name : 4,
        AssetTypes.transmission_poles.name : 0,
        AssetTypes.distribution_poles.name : 0,
        AssetTypes.transmission_overhead_lines.name : 100,
        AssetTypes.distribution_overhead_lines.name : 30,
    }
    
    assets = {
        AssetTypes.substation.name : {},
        AssetTypes.solar_panels.name : {},
        AssetTypes.buried_lines.name : {},
        AssetTypes.wind_turbines.name : {},
        AssetTypes.battery_storage.name :{},
        AssetTypes.transmission_poles.name : {},
        AssetTypes.distribution_poles.name : {},
        AssetTypes.transmission_overhead_lines.name : {},
        AssetTypes.distribution_overhead_lines.name : {},
    }
    
    for asset_type, probability in asset_probabilities.items():
        asset_id = 0
        for x1 in x:
            for y1 in y:    
                if random() < probability:
                    assets[asset_type.name][f"{asset_type.name} {asset_id}"] = {"coordinates" : (x1, y1), "heights_ft": heights_ft[asset_type.name]}
                    asset_id += 1
    
    p1 = Point(x.min(), y.min())
    p2 = Point(x.max(), y.min())
    p3 = Point(x.max(), y.max())
    p4 = Point(x.min(), y.max())
    pointList = [p1, p2, p3, p4, p1]
    poly = Polygon(pointList)
    mypoly = MultiPolygon([poly]) 
    
    return assets, mypoly


""" Module for handling computation of equity and resilience metrics.

Idea is to create separate function for computing each metric. 
"""

from pathlib import Path
from typing import Callable, Union, Dict, List
import json

from neo4j import GraphDatabase
import pandas as pd
import numpy as np

from erad.utils import util
from erad import exceptions


def validate_export_path(file_path: Union[str, Path], file_type: str):
    """Function for validating the export file path.

    Args:
        file_path (Union[str, Path]): Export file path
        file_type (str): File type to be exported e.g. .csv
    """

    output_path = Path(file_path)
    util.path_validation(output_path.parent)
    if output_path.suffix != file_type:
        raise exceptions.InvalidFileTypePassed(output_path, file_type)


def is_customer_getting_power(
    driver: GraphDatabase.driver, output_csv_path: str,
    load_list: List[str] = None
):

    """Function for checking whether customer is still connected
    to substation or not.

    Args:
        driver (GraphDatabase.driver): Instance of `GraphDatabase.driver`
            instance
        output_csv_path (str): CSV file path for exporting the metric.
    """
    if not load_list:
        load_list = []

    validate_export_path(output_csv_path, ".csv")
    cypher_query = """
                    MATCH (b:Bus)
                    MATCH (s:Substation)
                    WHERE b <> s
                    WITH b, shortestPath((b)-[:CONNECTS_TO*]-(s)) as p
                    MATCH (c:Load)-[:CONSUMES_POWER_FROM]-(b)
                    WITH c, p, apoc.coll.min([r in relationships(p) | r.survive]) AS max_p
                    RETURN c.name, max_p
                """
    metric_container = {"load_name": [], "metric": []}
    with driver.session() as session:
        result = session.read_transaction(
            lambda tx: tx.run(cypher_query).data()
        )

        for item in result:
            metric_container["load_name"].append(item["c.name"])
            metric_container["metric"].append(item["max_p"] \
                                              if item["c.name"] not in load_list else 1)

    df = pd.DataFrame(metric_container)
    df.to_csv(output_csv_path)


def energy_resilience_by_customer(
    driver: GraphDatabase.driver, output_csv_path: str,
    critical_infras: List = ["Grocery", "Hospital", "Pharmacy"]
):
    """Function for checking whether customer is still connected
    to substation or not.

    Args:
        driver (GraphDatabase.driver): Instance of `GraphDatabase.driver`
            instance
        output_csv_path (str): CSV file path for exporting the metric.
    """

    validate_export_path(output_csv_path, ".csv")

    metric_container = {"load_name": [], "metric": [], "critical_service": []}
    with driver.session() as session:

        for cs in critical_infras:
            cypher_query = (
                f"""
                    MATCH (lo:Load)
                    MATCH (cs:{cs})
                    """
                + """
                    WITH lo,cs, 
                        point.distance(point({longitude: lo.longitude, latitude:lo.latitude}), 
                        point({longitude: cs.longitude, latitude:cs.latitude}))/1000 AS d
                    RETURN lo.name, sum(toInteger(toBoolean(cs.survive) OR toBoolean(cs.backup))/d) AS gamma
                    """
                    # count(d)/sum(d) AS gamma
                    # WHERE cs.survive = 1 
            )
            result = session.read_transaction(
                lambda tx: tx.run(cypher_query).data()
            )

            for item in result:
                metric_container["load_name"].append(item["lo.name"])
                metric_container["metric"].append(item["gamma"])
                metric_container["critical_service"].append(cs)

    df = pd.DataFrame(metric_container)
    df.to_csv(output_csv_path)


def equity_based_energy_resilience_by_income(
    driver: GraphDatabase.driver,
    path_to_energy_resilience_metric: str,
    output_json_path: str,
    category: Dict[str, Callable] = {
        "low": lambda x: x < 90000,
        "medium": lambda x: (90000 < x < 110000),
        "high": lambda x: x > 110000,
    },
):
    """Function to compute the equity based energy resilience metric.

    Args:
        driver (GraphDatabase.driver): Instance of `GraphDatabase.driver`
            instance
        path_to_energy_resilience_metric (str): Path to energy resilience metric.
        output_json_path (str): JSON file path for exporting the metric.
        category (Dict[str, Callable]): Income categories
    """

    validate_export_path(output_json_path, ".json")
    util.path_validation(path_to_energy_resilience_metric)

    metric_container = {}

    with driver.session() as session:

        cypher_query = """
                            MATCH (lo:Load)
                            RETURN lo.income as income, lo.name as name
                        """
        result = session.read_transaction(
            lambda tx: tx.run(cypher_query).data()
        )

    resilience_metric = pd.read_csv(path_to_energy_resilience_metric)
    gamma_dict = (
        resilience_metric.groupby("load_name").sum()["metric"].to_dict()
    )

    for id, func in category.items():

        metric, income_flag_sum = 0, 0
        for load in result:
            load_income_flag = func(load["income"])
            metric += load_income_flag * gamma_dict[load["name"]]
            income_flag_sum += load_income_flag

        metric_container[id] = (
            metric / income_flag_sum if income_flag_sum else None
        )

    metric_values = [val for _, val in metric_container.items() if val]
    if metric_values:
        metric_container["community_resilience_equity_score"] = np.mean(
            metric_values
        ) / np.std(metric_values)

    with open(output_json_path, "w") as fpointer:
        json.dump(metric_container, fpointer)



from shapely.geometry import MultiPolygon, Point, LineString
from shapely.ops import nearest_points
import matplotlib.pyplot as plt
import scipy.stats as stats
import geopy.distance
import numpy as np
import stateplane


class GeoUtilities:
    
    @property
    def identify_stateplane_projection(self) -> str:
        """ Automatically identifies stateplane projection ID  """ 
        x = self.centroid.x
        y = self.centroid.y
        return stateplane.identify(x, y)
    
    def in_polygon(self, point : Point) -> bool:
        return self.multipolygon.contains(point)

    def distance_from_boundary(self, point : Point) -> float:
        """ Calculates distance of a point to polygon boundary. Correct calculations require conversion to cartesian coordinates""" 
        if self.multipolygon.contains(point):
            p1, p2 = nearest_points(self.boundary, point)
        else:
            p1, p2 = nearest_points(self.multipolygon, point)    
        coords_1 = (p1.y, p1.x)
        coords_2 = (p2.y, p2.x)
        return geopy.distance.geodesic(coords_1, coords_2).km    

    def distance_from_centroid(self, point : Point):
        """ Calculates distance of a point to polygon centroid. Correct calculations require conversion to cartesian coordinates """ 
        coords_1 = (self.centroid.y, self.centroid.x)
        coords_2 = (point.y, point.x)
        return geopy.distance.geodesic(coords_1, coords_2).km
    


class ProbabilityFunctionBuilder:
    """Class containing utility fuctions for sceario definations."""
    
    
    def __init__(self, dist, params):
        """Constructor for BaseScenario class.

        Args:
            dist (str): Name of teh distribution. Should follow Scipy naming convention
            params (list): A list of parameters for the chosen distribution function. See Scipy.stats documentation
        """
        
        self.dist = getattr(stats, dist)
        self.params = params
        return 

    def sample(self):
        """Sample the distribution """
        return self.dist.rvs(*self.params, size=1)[0]

    def plot_cdf(self, x:np.linspace, ax =None, label="") -> None:
        """Plot the cumalative distribution fuction"""
        cdf = self.dist.cdf
        if ax is None:
            plt.plot(x,cdf(x, *self.params), label=label)
        else:
            ax.plot(x,cdf(x, *self.params), label=label)
    

    def probability(self, value: float) -> float:
        """Calculates survival probability of a given asset.

        Args:
            value (float): value for vetor of interest. Will change with scenarions
        """
        cdf = self.dist.cdf
        return cdf(value, *self.params)


import json
from typing import List

import overpass
import polars

# Define your polygon coordinates as a string

GROCERY_TAGS = [
    'shop=supermarket', 'shop=grocery', 'shop=convenience',
    'shop=market', 'shop=healthfood', 'shop=organic'
]

HOSPITAL_TAGS = [
        'amenity=hospital', 'healthcare=hospital', 'building=hospital',
        'amenity=clinic', 'healthcare=centre','healthcare=pharmacy',
]


def export_json(json_parcels, out_path):
    with open(out_path, 'w') as fp:
        json.dump(json_parcels, fp)

def export_csv(json_parcels, out_path:str):

    csv_data = {
        "id": [],
        "longitude": [],
        "latitude": [],
    }

    for feature in json_parcels['features']:
        for key in  feature['properties']:
            if key not in csv_data:
                csv_data[key] = []
    
    
    for feature in json_parcels['features']:
        csv_data["id"].append(feature["id"])
        csv_data["longitude"].append(feature["geometry"]["coordinates"][0])
        csv_data["latitude"].append(feature["geometry"]["coordinates"][1])

        for key in csv_data:
            if key not in ["id", "longitude", "latitude"]:
                csv_data[key].append(feature['properties'].get(key, None))

    df = polars.from_dict(csv_data)
    df.write_csv(out_path)


def get_sites(polygon: str, tags: List[str] ):
    # polygon = "33.6812,-118.5966, 34.3407,-118.1390" latitude, longitude 
    json_parcels = {
        "type": "FeatureCollection",
        "features": []
    }

    for tag in tags:

        query = f"""
        node[{tag}]({polygon});
        out;
        """

        # Create an instance of the Overpass API
        api = overpass.API()

        # Send the query and retrieve the data
        response = api.Get(query)

        json_parcels['features'].extend(response['features'])
        print(f'Number of parcels extracted of shop type {tag}: ', len(response['features']))

    print('Number of parcels extracted: ', len(json_parcels['features']))
    return json_parcels




""" Module contains the base class defination for all scenarios.

Each scenario type will extend the base class to implement a given scenario
"""


from shapely.geometry import MultiPolygon, Polygon, LineString, Point
from datetime import datetime
from typing import *
import shapely

from erad.scenarios.common import AssetTypes

class BaseScenario:
    
    """Base class for scenario defination.

    Attributes:
        geodata (MultiPolygon, Point, LineString): Region of interest
        probability_model (dict): Dictionary mapping asset types to probability funcitons
        timestamp (datetime): Scenario occurance time 
        kwargs (dict): Additional parameters relevant for a particular scenario type
        
    """
    
    def __init__(self,  geodata : Union[MultiPolygon, Point, LineString] , probability_model : dict, timestamp : datetime, **kwargs) -> None:
        """Constructor for BaseScenario class.

        Args:
            geodata (MultiPolygon, Point, LineString): Region of interest
            probability_model (dict): Dictionary mapping asset types to probability funcitons
            timestamp (datetime): Scenario occurance time 
            kwargs (dict): Additional parameters relevant for a particular scenario type
        """
        if probability_model is None:
            probability_model = self.fragility_curves
        self.valitate_user_defined_fragility_curves(probability_model)
        
    
        if isinstance(geodata, Polygon):
           geodata = MultiPolygon([geodata]) 
        
        if isinstance(geodata, MultiPolygon):
            print("Is multipolygon")
            self.multipolygon = geodata
        elif isinstance(geodata, Point):
            print("Is point")
            self.origin = geodata
        elif isinstance(geodata, LineString):
            print("Is linestring")
            self.front = geodata
        else:
            print(geodata, Point)
            raise Exception(f"Invalid data type {type(geodata)}")
        
        self.probability_model = probability_model
        self.to_projection = f"epsg:{self.identify_stateplane_projection}"
        self.timestamp = timestamp
        return
    
    @property
    def area(self) -> float:
        """Method to calculate area of affected region."""
        raise NotImplementedError("Method needs to be defined in derived classes")
    
    @property
    def polygon(self) -> MultiPolygon:
        """Method to return polygon for the affected region."""
        raise NotImplementedError("Method needs to be defined in derived classes")
        
    @property
    def boundary(self) -> LineString:
        """Method to return boundary for the affected region."""
        raise NotImplementedError("Method needs to be defined in derived classes")
        
    @property
    def centroid(self) -> Point:
        """Method to return the centroid of the affected region."""
        raise NotImplementedError("Method needs to be defined in derived classes")
    
    def increment_time(self) -> dict:
        """Method to increment simulation time for time evolviong scenarios."""
        raise NotImplementedError("Method needs to be defined in derived classes")

    def calculate_survival_probability(self, assets : dict, timestamp : datetime) -> dict:
        """Method to calculate survival probaility of asset types.

        Args:
            assets (dict): The dictionary of all assets and their corresponding asset types
        """
        raise NotImplementedError("Method needs to be defined in derived classes")

    def plot(self):
        """Method to plot survival probaility of in the region of interest"""
        raise NotImplementedError("Method needs to be defined in derived classes")
   
    def asset_survial_probability(self, asset_type):
        raise NotImplementedError("Method needs to be defined in derived classes")
    
    def valitate_user_defined_fragility_curves(self, distributions):
        for asset_type in distributions:
            assert AssetTypes.has_asset(asset_type), f"{asset_type} is not a valid asset type. Valid options are {list(AssetTypes.__members__.keys())}"
        return


""" This module contains utility functions 
utilized throughout the modules and subpackages
contained within db subpackage."""

from neo4j import GraphDatabase

def _run_read_query(driver:GraphDatabase.driver, cypher_query: str):
    """ Runs a cypher query and returns result. """

    with driver.session() as session:
        result = session.read_transaction(
            lambda tx: tx.run(cypher_query).data()
        )
    return result


"""
This module will update the survive property based on whether 
the critica infrastructure is going to be powered from microgrid or not.
"""


from neo4j import GraphDatabase

from erad.metrics.check_microgrid import check_for_microgrid


def apply_microgrid_to_critical_infra(
        driver: GraphDatabase.driver,
        factor: float= (0.5 * 0.4)
):
    """ Function that will update the survive property of 
    critical infrastructure if it can get power from microgrid."""

    islands_jsons = check_for_microgrid(driver, output_json_path=None)
    infra_island = [subdict['sinks'] for _, subdict \
                      in islands_jsons.items() if len(subdict['sources'])>0]
    infra_microgrid = [1 if subdict['source_capacity'] >= subdict['sink_capacity']*factor  else 0 for _, subdict \
                      in islands_jsons.items()]

    infra_survives = [el for id, el in enumerate(infra_island) \
                      if infra_microgrid[id]]
    all_sinks = [x for el in infra_survives for x in el]
    infra_survives = [x for el in infra_survives for x in el if 'load' not in x]
    
    for infra in infra_survives:
        cypher_query = f"""
                        MATCH (c)
                        WHERE c.name = $cname
                        SET c.survive = 1
                        SET c.survival_probability = 1
                    """

        with driver.session() as session:
                session.write_transaction(
                    lambda tx: tx.run(
                        cypher_query,
                        cname=infra,
                    )
                )
    return all_sinks


""" Module for extracting assets from OpenDSS model.

Examples:

    >>> from erad import utils
    >>> extract_opendss_model(
            <path_to_opendss_master_file>,
            <output_folder>
        )
"""

# standard imports
from genericpath import exists
from pathlib import Path
import logging
from typing import List

# third-party imports
import opendssdirect as dss
import pandas as pd
import networkx as nx
from shapely.geometry import MultiPoint
import stateplane

# internal imports
from erad.utils.util import path_validation, setup_logging
from erad.exceptions import OpenDSSCommandError, MultiStatePlaneError


logger = logging.getLogger(__name__)


def get_transformers(dss_instance: dss) -> List:
    """Function to return list of transformers in opendss models.

    Args:
        dss_instance (dss): OpenDSS instance with models preloaded

    Returns:
        List: List of transformer metadata object
    """

    transformer_container = []
    flag = dss_instance.Transformers.First()
    while flag > 0:
        trans_name = dss_instance.CktElement.Name().lower()
        buses = dss_instance.CktElement.BusNames()
        bus1, bus2 = buses[0].split(".")[0], buses[1].split(".")[0]

        transformer_container.append(
            {
                "name": trans_name,
                "type": "Transformer",
                "source": bus1,
                "target": bus2,
                "kva": dss_instance.Transformers.kVA(),
                "num_phase": dss_instance.CktElement.NumPhases(),
            }
        )
        flag = dss_instance.Transformers.Next()
    return transformer_container


def get_line_sections(dss_instance: dss) -> List:
    """Function to return list of all line segments in opendss model.

    Args:
        dss_instance (dss): OpenDSS instance with models preloaded

    Returns:
        List: List of line segment metadata object
    """
    UNIT_MAPPER = {
        0: 0,
        1: 1.60934,
        2: 0.3048,
        3: 1,
        4: 0.001,
        5: 0.0003048,
        6: 0.0000254,
        7: 0.00001,
    }

    sections_container = []
    flag = dss_instance.Lines.First()
    while flag > 0:
        section_name = dss_instance.CktElement.Name().lower()
        buses = dss_instance.CktElement.BusNames()
        bus1, bus2 = buses[0].split(".")[0], buses[1].split(".")[0]

        sections_container.append(
            {
                "name": section_name,
                "type": "LineSegment",
                "source": bus1,
                "target": bus2,
                "length_km": UNIT_MAPPER[dss_instance.Lines.Units()]
                * dss_instance.Lines.Length(),
                "ampacity": dss_instance.Lines.NormAmps(),
                "num_phase": dss_instance.CktElement.NumPhases(),
            }
        )

        flag = dss_instance.Lines.Next()
    return sections_container


def get_buses(dss_instance: dss) -> List:
    """Function to return list of all buses in opendss model.

    Args:
        dss_instance (dss): OpenDSS instance with models preloaded

    Returns:
        List: List of bus metadata object
    """
    buses_container = []
    for bus in dss_instance.Circuit.AllBusNames():
        dss_instance.Circuit.SetActiveBus(bus)
        buses_container.append(
            {
                "name": bus,
                "type": "Bus",
                "kv": dss_instance.Bus.kVBase(),
                "longitude": dss_instance.Bus.X(),
                "latitude": dss_instance.Bus.Y(),
            }
        )
    return buses_container


def get_capacitors(dss_instance: dss) -> List:
    """Function to return list of all capacitors in opendss model.

    Args:
        dss_instance (dss): OpenDSS instance with models preloaded

    Returns:
        List: List of capacitor metadata object
    """

    capacitors_container = []
    flag = dss_instance.Capacitors.First()
    while flag > 0:
        capacitor_name = dss_instance.CktElement.Name().lower()
        buses = dss_instance.CktElement.BusNames()
        bus1 = buses[0].split(".")[0]

        capacitors_container.append(
            {
                "name": capacitor_name,
                "type": "Capacitor",
                "source": bus1,
                "kv": dss_instance.Capacitors.kV(),
                "kvar": dss_instance.Capacitors.kvar(),
            }
        )

        flag = dss_instance.Capacitors.Next()
    return capacitors_container


def get_pvsystems(dss_instance: dss) -> List:
    """Function to return list of all pv systems in opendss model.

    Args:
        dss_instance (dss): OpenDSS instance with models preloaded

    Returns:
        List: List of PVsystem metadata object
    """

    pvs_container = []
    flag = dss_instance.PVsystems.First()
    while flag > 0:
        pv_name = dss_instance.CktElement.Name().lower()
        buses = dss_instance.CktElement.BusNames()
        bus1 = buses[0].split(".")[0]

        pvs_container.append(
            {
                "name": pv_name,
                "type": "PVSystem",
                "source": bus1,
                "rated_power": dss_instance.PVSystems.Pmpp(),
            }
        )

        flag = dss_instance.PVsystems.Next()
    return pvs_container


def get_loads(dss_instance):
    """Function to return list of all loads in opendss model.

    Args:
        dss_instance (dss): OpenDSS instance with models preloaded

    Returns:
        List: List of load metadata object
    """

    loads_container = []
    flag = dss_instance.Loads.First()
    while flag > 0:
        load_name = dss_instance.CktElement.Name().lower()
        buses = dss_instance.CktElement.BusNames()
        bus1 = buses[0].split(".")[0]

        loads_container.append(
            {
                "name": load_name,
                "type": "Load",
                "source": bus1,
                "kw": dss_instance.Loads.kW(),
                "kvar": dss_instance.Loads.kvar(),
            }
        )

        flag = dss_instance.Loads.Next()
    return loads_container


def execute_dss_command(dss_instance: dss, dss_command: str) -> None:
    """Pass the valid dss command to be executed.

    Args:
        dss_instance (dss): OpenDSS instance with models preloaded
        dss_command (str): DSS command sring to be executed

    Raises:
        OpenDSSCommandError: Raises this because opendss command execution
            ran into error

    """
    error = dss_instance.run_command(dss_command)
    if error:
        logger.error(f"Error executing command {dss_command} >> {error}")
        raise OpenDSSCommandError(
            f"Error executing command {dss_command} >> {error}"
        )
    logger.info(f"Sucessfully executed the command, {dss_command}")


def get_bounding_box(master_file: str, buffer: float = 1000) -> List:
    """Creates a bounding box coordinate for covering region of opendss model.

    Args:
        master_file (str): Path to master dss file
        buffer (float): Buffer distance around distribution model in meter

    Raises:
        MultiStatePlaneError: Raises this if opendss models lies in multiple
            state plane coordinates

    Returns:
        List: List of bounding box coordinates (lower_left, upper_right)
    """

    # Get bounding box for opendss network
    # Do a basic check on the path
    master_file = Path(master_file)
    path_validation(master_file)
    logger.debug(f"Attempting to read case file >> {master_file}")

    # Clear memory and compile dss file
    dss.run_command("Clear")
    dss.Basic.ClearAll()
    execute_dss_command(dss, f"Redirect {master_file}")

    # Get all the points
    points = []
    for bus in dss.Circuit.AllBusNames():
        dss.Circuit.SetActiveBus(bus)
        points.append([dss.Bus.X(), dss.Bus.Y()])

    # Create a multipoint to get bounds
    multi_points = MultiPoint(points)
    bounds = multi_points.bounds

    # Get EPSG value for converting into coordinate reference system
    if stateplane.identify(bounds[0], bounds[1]) != stateplane.identify(
        bounds[2], bounds[3]
    ):
        raise MultiStatePlaneError(
            f"The regions uses multiple stateplane coordinate system"
        )

    epsg_value = stateplane.identify(bounds[0], bounds[1])

    # Let's project all the WGS84 coordinates into
    # transformed coordinates this will make sure distance is in meter
    transformed_points = [
        stateplane.from_lonlat(*point, epsg_value) for point in points
    ]

    # Create a multipoint from the transformed coordinates
    transformed_multipoint = MultiPoint(transformed_points).buffer(buffer)

    # Get the bounds and convert back to wsg84 format
    transformed_bounds = transformed_multipoint.bounds
    bounds_wsg84 = stateplane.to_lonlat(
        transformed_bounds[0], transformed_bounds[1], epsg_value
    ) + stateplane.to_lonlat(
        transformed_bounds[2], transformed_bounds[3], epsg_value
    )

    return bounds_wsg84


def extract_export_opendss_model(
    master_file: str, output_folder_path: str
) -> None:
    """Extract the opendss models and exports into csv file format.

    Args:
        master_file (str): Path to opendss master file
        output_folder_path (str): Folder path for exporting the models to.
    """

    # Do a basic check on the path
    master_file = Path(master_file)
    path_validation(master_file)
    logger.debug(f"Attempting to read case file >> {master_file}")

    # Clear memory and compile dss file
    dss.run_command("Clear")
    dss.Basic.ClearAll()
    execute_dss_command(dss, f"Redirect {master_file}")

    # Initial container
    transformers = get_transformers(dss)
    line_sections = get_line_sections(dss)
    buses = get_buses(dss)
    capacitors = get_capacitors(dss)
    pv_systems = get_pvsystems(dss)
    loads = get_loads(dss)

    output_folder_path = Path(output_folder_path)
    output_folder_path.mkdir(exist_ok=True)
    transformers_df = pd.DataFrame(transformers)
    transformers_df.to_csv(output_folder_path / "transformers.csv")

    line_sections_df = pd.DataFrame(line_sections)
    line_sections_df.to_csv(output_folder_path / "line_sections.csv")

    buses_df = pd.DataFrame(buses)
    buses_df.to_csv(output_folder_path / "buses.csv")

    capacitors_df = pd.DataFrame(capacitors)
    capacitors_df.to_csv(output_folder_path / "capacitors.csv")

    pv_systems_df = pd.DataFrame(pv_systems)
    pv_systems_df.to_csv(output_folder_path / "pv_systems.csv")

    loads_df = pd.DataFrame(loads)
    loads_df.to_csv(output_folder_path / "loads.csv")



""" Module for injecting flooding scenario data into graph database. """


from neo4j import GraphDatabase
from shapely import MultiPolygon, Polygon

from erad.db import disaster_input_model
from erad.scenarios import flood_scenario
from erad.db.assets.distribution_lines import (
    _update_distribution_overhead_lines,
    _update_distribution_underground_lines
)

from erad.db.assets.critical_infras import (
    _update_critical_infra_based_on_grid_access_fast
)


def inject_polygon_flooding(
    scenario_input: disaster_input_model.PolygonFlooding,
    driver: GraphDatabase.driver,
    critical_infras: None
):
    """Function to update the survival probability of different assets
    due to user input based flooding scenario.

    Args:
        scenario_input (scenario_model.PolygonFlooding): Should be an instance
            of `scenario_model.PolygonFlooding` data model.
        driver (GraphDatabase.driver): Instance of `GraphDatabase.driver`
            instance
    """
    if not critical_infras:
        critical_infras = []


    scenario = flood_scenario.FlooadScenario(
        MultiPolygon(
            [Polygon(el) for el in scenario_input.polygon]
        ),
        scenario_input.probability_model,
        scenario_input.timestamp,
        file_flow=scenario_input.file_flow,
        file_levels=scenario_input.file_levels,
        file_gaugues=scenario_input.file_gaugues
    )
    
    _update_distribution_underground_lines(scenario, driver, scenario_input.timestamp)
    _update_distribution_overhead_lines(scenario, driver, scenario_input.timestamp)
    if len(critical_infras):
        _update_critical_infra_based_on_grid_access_fast(critical_infras, driver)
    
            




import matplotlib.pyplot as plt
from erad.scenarios.utilities import ProbabilityFunctionBuilder, GeoUtilities
from erad.constants import EARTHQUAKE_HISTORIC_CSV_PATH, DATA_FOLDER
from shapely.geometry import MultiPolygon, Point, LineString
from erad.scenarios.abstract_scenario import BaseScenario
from erad.scenarios.common import asset_list
from datetime import datetime
import pandas as pd
import numpy as np
import math
import os

from erad.scenarios.common import AssetTypes
from erad.scenarios.utilities import ProbabilityFunctionBuilder

class EarthquakeScenario(BaseScenario, GeoUtilities):
    """Base class for EarthquakeScenario. Extends BaseScenario and GeoUtilities

    Attributes:
        origin (Point): Earthquake origin point
        probability_model (dict): Dictionary mapping asset types to probability funcitons
        timestamp (datetime): Scenario occurance time 
        kwargs (dict): Additional parameters relevant for a particular scenario type      
    """
    fragility_curves = {
            # Hierarchical seismic vulnerability assessment of power transmission systems: sensitivity analysis of fragility curves and clustering algorithms
            AssetTypes.substation.name : ProbabilityFunctionBuilder("norm", [0.75, 0.20]),
            # AssetTypes.solar_panels.name : ProbabilityFunctionBuilder("norm", [85, 10]),
            # Seismic performance of buried electrical cables: evidence-based repair rates and fragility functions
            # SEISMIC FRAGILITY OF UNDERGROUND ELECTRICAL CABLES IN THE 2010-11 CANTERBURY (NZ) EARTHQUAKES
            # AssetTypes.buried_lines.name : ProbabilityFunctionBuilder("uniform", [0, 1]),
            # Seismic Fragility Analysis of Monopile Offshore Wind Turbines under Different Operational Conditions
            AssetTypes.wind_turbines.name : ProbabilityFunctionBuilder("norm", [0.62, 0.17]),        
            # AssetTypes.battery_storage.name : ProbabilityFunctionBuilder("norm", [85, 10]),
            # Multi-hazard typhoon and earthquake collapse fragility models for transmission towers: An active learning reliability approach using gradient boosting classifiers
            AssetTypes.transmission_poles.name  :  ProbabilityFunctionBuilder("norm", [1.5, 0.45]),
            # Seismic performance and fragility analysis of power distribution concrete poles
            # AssetTypes.distribution_poles.name  : ProbabilityFunctionBuilder("norm", [0.73, 0.15]),
            #Improving the resilience of distribution network in coming across seismic damage using mobile battery energy storage system
            AssetTypes.distribution_poles.name  : ProbabilityFunctionBuilder("norm", [0.6, 0.12]),
            # https://www.researchgate.net/figure/Fragility-curves-that-show-the-probability-of-electric-power-lines-being-in-each-failure_fig3_339017910
            AssetTypes.transmission_overhead_lines.name  : ProbabilityFunctionBuilder("norm", [0.72, 0.1]),   
            # https://www.fema.gov/sites/default/files/2020-10/fema_hazus_earthquake_technical_manual_4-2.pdf
            AssetTypes.distribution_overhead_lines.name  : ProbabilityFunctionBuilder("lognorm", [0.3, 0.25,0.4]),
            
        }
    
    intensity_map = LineString([[1, 2.5], [2, 3.0], [3, 3.5], [4, 4.0], [5, 4.5], [6, 5.0], [7, 5.5], [8, 6.0], [9, 6.5], [10, 7.0], [11, 7.5], [12, 8.0]])
    def __init__(self,  origin : Point , probability_model : dict, timestamp : datetime, **kwargs) -> None:
        """Constructor for EarthquakeScenario.

        Args:
            origin (Point): Earthquake origin point
            probability_model (dict): Dictionary mapping asset types to probability funcitons
            timestamp (datetime): Scenario occurance time 
            kwargs (dict): Additional parameters relevant for a particular scenario type      
        """
        
        super(EarthquakeScenario, self).__init__(origin, probability_model, timestamp, **kwargs)
        self.kwargs = kwargs
        return
    
    @classmethod
    def from_dynamic_model(cls, origin : Point , probability_function : dict, timestamp : datetime):
        raise NotImplementedError("Method needs to be defined in derived classes")
    
    @classmethod
    def from_historical_earthquake_by_code(cls, earthquake_code : str, probability_function : dict= None):
        """Class method for EarthquakeScenario.

        Args:
            earthquake_code (str): Code for a historic eqrthquake event
            probability_function (dict): Dictionary mapping asset types to probability funcitons
        """
        
        data_file = os.path.join(DATA_FOLDER, EARTHQUAKE_HISTORIC_CSV_PATH)
        assert os.path.exists(data_file), f"The data file {data_file} not found"
        earthquake_data = pd.read_csv(data_file, index_col=None)
        earthquake_data['DateTime'] = pd.to_datetime(earthquake_data['Date'] + ' ' + earthquake_data['Time'], format='%m/%d/%Y %H:%M:%S')
        earthquake_data_filtered = earthquake_data[earthquake_data['ID'] == earthquake_code]
        
        timestamp = earthquake_data_filtered.DateTime.values[0]
        kwargs = {
            "Magnitude": earthquake_data_filtered.Magnitude.values[0],
            "Depth" : earthquake_data_filtered.Depth.values[0],
            }
        long = earthquake_data_filtered.Longitude.values[0]
        lat = earthquake_data_filtered.Latitude.values[0]
        origin = Point(long, lat) 
        return cls(origin, probability_function, timestamp, **kwargs)
    
    @property
    def centroid(self):
        """Method to return the centroid of the affected region."""
        return self.origin
    
    def increment_time(self):
        """Method to increment simulation time for time evolviong scenarios."""
        raise NotImplementedError("Method needs to be defined in derived classes")

    def calculate_survival_probability(self, assets : dict, timestamp : datetime) -> dict:
        """Method to calculate survival probaility of asset types.

        Args:
            assets (dict): The dictionary of all assets and their corresponding asset types
        """
        for asset_type, asset_dict in assets.items():
            # assert asset_type in self.probability_model, f"Survival probability for asset type '{asset_type}' not found in the passed probability_model"
            if asset_type in self.probability_model:
                
                probability_function = self.probability_model[asset_type]
                for asset_name, asset_ppty in asset_dict.items():
                    coords = Point(asset_ppty["coordinates"][::-1])
                    
                    epicenter_distance = self.distance_from_centroid(coords)
        
                    depth = self.kwargs["Depth"]
                    magnitude = self.kwargs["Magnitude"]

                    # Valutazione speditiva di sicurezza sismica degli edifici esistenti
                    hypocentral_distance = (depth**2 + epicenter_distance**2)**0.5
                    l = LineString([[0, magnitude], [100, magnitude]])
                    p = self.intensity_map.intersection(l)
                    Imcs = p.xy[0][0]
                    intensity = Imcs + 3 - 4.3 * math.log10(hypocentral_distance)
                    intensity = 0 if intensity < 0 else intensity
                    pga = 10**((intensity /3) - 1) / 9.81
                    probility =  1- probability_function.probability(pga)
                    assets[asset_type][asset_name]["survival_probability"] = probility
        return assets

    def plot(self, d : float):
        """Method to plot survival probaility of in the region of interest"""
        m = range(2, 10)
        h = np.linspace(14, 70, 100)
       
        
        fig = plt.figure()
        ax = fig.add_subplot(111)
        for magnitude in m:
            points = []
            for depth in h:
                hypocentral_distance = (depth**2 + d**2)**0.5
                intensity = 0.66* magnitude - 1.13 * np.log(hypocentral_distance) -0.0072 * hypocentral_distance + 3.73
                act_int = 10**intensity
                survival_prob = 1 - act_int / 383641.228284058
                if  survival_prob > 1:
                    survival_prob = 1
                elif survival_prob < 0:
                    survival_prob = 0
                
                points.append(survival_prob)
                
            print(points[0])
            ax.plot(h, points, label=f"magnitude: {magnitude}")
            ax.legend()
        plt.show()
            
            




""" Module for injecting scenario data into graph database. """


from neo4j import GraphDatabase
import shapely

from erad.db import disaster_input_model
from erad.scenarios import earthquake_scenario
from erad.db.assets.distribution_lines import (
    _update_distribution_overhead_lines,
    _update_distribution_underground_lines
)

from erad.db.assets.critical_infras import (
    _update_critical_infra_based_on_grid_access_fast
)


def inject_point_earthquake(
    scenario_input: disaster_input_model.PointEarthquake,
    driver: GraphDatabase.driver,
    critical_infras: None
):
    """Function to update the survival probability of different assets
    due to user input based earthquake scenario.

    Args:
        scenario_input (scenario_model.PointEarthquake): Should be an instance
            of `scenario_model.PointEarthquake` data model.
        driver (GraphDatabase.driver): Instance of `GraphDatabase.driver`
            instance
    """
    if not critical_infras:
        critical_infras = []


    scenario = earthquake_scenario.EarthquakeScenario(
        shapely.geometry.Point(
            scenario_input.longitude, scenario_input.latitude
        ),
        scenario_input.probability_model,
        scenario_input.timestamp,
        Magnitude=scenario_input.magnitude,
        Depth=scenario_input.depth,
    )
    
    _update_distribution_overhead_lines(scenario, driver, scenario_input.timestamp)
    _update_distribution_underground_lines(scenario, driver, scenario_input.timestamp)
    if len(critical_infras):
        _update_critical_infra_based_on_grid_access_fast(critical_infras, driver)
    
            



""" This module will update the backup property for 
critical infrastructure. """

# standard imports
from typing import List
import random

# third party imports
from neo4j import GraphDatabase

# internal imports
from erad.db.utils import _run_read_query

def apply_backup_program(
    driver: GraphDatabase.driver,
    electricity_backup: any,
    critical_infras: List[str],
):

    """ Function that will update the backup 
    property of critical infras based on backup percentage.
    
    Args:
        driver (GraphDatabase.driver): Neo4J Driver instance
        electricity_backup (float): backup percentage number between 0 and 1 or list of infras to set as backup
        critical_infras (List[str]): list of critical infrastructure
    """

    infra_with_backups = []
    for cri_infra in critical_infras:
        cypher_query = f"""
                        MATCH (c:{cri_infra})
                        RETURN c.longitude, c.latitude, c.name, c.backup
                    """
        infras = _run_read_query(driver, cypher_query)
        
        for infra in infras:
            
            cypher_write_query = f"""
                MATCH (c:{cri_infra})
                WHERE c.name = $cname
                SET c.backup = $backup
            """

            if not isinstance(electricity_backup, list):
                backup = 1 if random.random() < electricity_backup else 0
            else:
                backup = 1 if infra['c.name'] in electricity_backup else 0 

            if backup == 1:
                infra_with_backups.append(infra['c.name'])

            with driver.session() as session:
                session.write_transaction(
                    lambda tx: tx.run(
                        cypher_write_query,
                        cname=infra['c.name'],
                        backup= backup
                    )
                )

    return infra_with_backups


""" Utility functions for dealing with SMART DS dataset. 

Examples:

    >>> from erad import ditto_utils
    >>> ditto_utils.download_smartds_data('P4R',  '.')

"""

# standard libraries
from pathlib import Path
import shutil
import logging
from typing import List

# third-party libraries
import boto3
from botocore import UNSIGNED
from botocore.config import Config
from ditto.store import Store
from ditto.readers.opendss.read import Reader
from ditto.network.network import Network
from ditto.models.power_source import PowerSource
import networkx as nx
from networkx.readwrite import json_graph


# internal libraries
from erad.constants import SMARTDS_VALID_AREAS, SMARTDS_VALID_YEARS
from erad.exceptions import SMARTDSInvalidInput, DittoException
from erad.utils.util import timeit, write_file, path_validation
from erad.utils.util import read_file, setup_logging


logger = logging.getLogger(__name__)


@timeit
def download_aws_dir(
    bucket: str, path: str, target: str, unsigned=True, **kwargs
) -> None:
    """Utility function download data from AWS S3 directory.

    Args:
        bucket (str): Name of the bucket.
        path (str): S3 bucket prefix
        target (str): Path for downloading the data
        unsigned (bool): Indicate whether to use credential or not
        kwargs (dict): Keyword arguments accepted by `boto3.client`
    """

    target = Path(target)
    if unsigned:
        client = boto3.client("s3", config=Config(signature_version=UNSIGNED))
    else:
        if kwargs:
            client = boto3.client("s3", **kwargs)
        else:
            client = boto3.client("s3")

    # Handle missing / at end of prefix
    if not path.endswith("/"):
        path += "/"

    paginator = client.get_paginator("list_objects_v2")
    for result in paginator.paginate(Bucket=bucket, Prefix=path):

        # Download each file individually
        for key in result["Contents"]:

            # Calculate relative path
            rel_path = key["Key"][len(path) :]

            # Skip paths ending in /
            if not key["Key"].endswith("/"):
                local_file_path = target / rel_path
                local_file_path.parent.mkdir(parents=True, exist_ok=True)
                client.download_file(bucket, key["Key"], str(local_file_path))


@timeit
def download_smartds_data(
    smartds_region: str,
    output_path: str = "./smart_ds_downloads",
    year: int = 2018,
    area: str = "SFO",
    s3_bucket_name: str = "oedi-data-lake",
    folder_name: str = "opendss_no_loadshapes",
    cache_folder: str = "cache",
) -> str:
    """Utility function to download SMARTDS data from AWS S3 bucket.

    Args:
        smartds_region (str): SMARTDS region name
        output_path (str): Path for downloaded data
        year (int): Valid year input for downloading the data
        area (str): Valid SMARTDS area
        s3_bucket_name (str): S3 bucket name storing the SMARTDS data
        folder_name (str): S3 bucket folder to download
        cache_folder (str): Folder path for caching the results

    Raises:
        SMARTDSInvalidInput: Raises this error if year and/or area
            provided is not valid.

    Returns:
        str: Folder path containing downloaded data.
    """
    if year not in SMARTDS_VALID_YEARS or area not in SMARTDS_VALID_AREAS:
        raise SMARTDSInvalidInput(
            f"Not valid input! year= {year} area={area}, \
            valid_years={SMARTDS_VALID_YEARS}, valid_areas={SMARTDS_VALID_AREAS}"
        )

    output_path = Path(output_path)
    cache_folder = Path(cache_folder)

    output_path.mkdir(exist_ok=True)
    cache_folder.mkdir(exist_ok=True)

    cache_key = (
        f"{smartds_region}__{year}__{area}__{s3_bucket_name}_{folder_name}"
    )
    cache_data_folder = cache_folder / cache_key
    output_folder = output_path / cache_key

    if cache_data_folder.exists():
        logger.info(f"Cache hit for {cache_data_folder}")
        shutil.copytree(cache_data_folder, output_folder, dirs_exist_ok=True)

    else:
        logger.info(
            f"Cache missed reaching to AWS for downloading the data ..."
        )
        output_folder.mkdir(exist_ok=True)
        prefix = f"SMART-DS/v1.0/{year}/{area}/{smartds_region}/scenarios/base_timeseries/{folder_name}/"
        download_aws_dir(s3_bucket_name, prefix, output_folder)
        shutil.copytree(output_folder, cache_data_folder, dirs_exist_ok=False)

    logger.info(f"Check the folder {output_folder} for downloaded data")
    return output_folder


@timeit
def _create_networkx_from_ditto(
    output_path: str, file_name: str, **kwargs
) -> List:
    """Creates networkx graph from OpenDSS model using Ditto.

    Args:
        output_path (str): Path to store the networkx
            data in json file format
        file_name (str): JSON file name used to export
            the network
        kwargs (dict): Keyword arguments accepted
            by Ditto

    Raises:
        DittoException: Raises if multiple sources are found.


    Returns:
        List: Pair of networkx graph and
            path containing JSON file
    """
    file_name = Path(file_name).stem
    logger.debug(
        "Attempting to create NetworkX representation from OpenDSS \
        files using DiTTo"
    )

    path_validation(output_path)

    store = Store()
    reader = Reader(
        master_file=kwargs["master_file"],
        buscoordinates_file=kwargs["buscoordinates_file"],
        coordinates_delimiter=kwargs["coordinates_delimiter"],
    )
    reader.parse(store)

    all_sources = []
    for i in store.models:
        if isinstance(i, PowerSource) and i.connecting_element is not None:
            all_sources.append(i)
        elif isinstance(i, PowerSource):
            print(
                "Warning - a PowerSource element has a None connecting element"
            )

    if len(all_sources) > 1:
        raise DittoException(
            f"This feeder has lots of sources {len(all_sources)}"
        )

    ditto_graph = Network()
    ditto_graph.build(store, all_sources[0].connecting_element)
    ditto_graph.set_attributes(store)

    data = dict(ditto_graph.graph.nodes.data())
    data_new = {}
    for node, node_data in data.items():
        try:
            data_new[node] = node_data["positions"][0]._trait_values
        except Exception as e:
            connecting_node = node_data["connecting_element"]
            data_new[node] = data[connecting_node]["positions"][0]._trait_values

    adj_file = file_name + ".adjlist"
    nx.write_adjlist(ditto_graph.graph, output_path / adj_file)
    g = nx.read_adjlist(output_path / adj_file)
    nx.set_node_attributes(g, data_new)

    data = json_graph.adjacency_data(g)
    json_file = file_name + ".json"
    output_file = output_path / json_file
    write_file(data, output_file)

    logger.debug(
        f"Successfully created json file representing the network \
        check the file {output_file}"
    )

    return (g, output_file)


def create_networkx_from_ditto(
    output_path: str, file_name: str, **kwargs
) -> None:
    """Creates networkx graph from OpenDSS model using Ditto.

    Args:
        output_path (str): Path to store the networkx
            data in json file format
        file_name (str): JSON file name used to export
            the network
        kwargs (dict): Keyword arguments accepted
            by Ditto
    """
    try:
        output_path = Path(output_path)
        return _create_networkx_from_ditto(output_path, file_name, **kwargs)
    finally:
        for file_path in output_path.iterdir():
            if file_path.suffix == ".adjlist":
                file_path.unlink(missing_ok=True)


def create_networkx_from_json(json_file_path: str):
    """Returns networkx graph from JSON file."""
    content = read_file(json_file_path)
    return json_graph.adjacency_graph(content)



""" Module for storing pydantic model for neo4j credential. """

from pydantic import BaseModel, validator

from erad.exceptions import DatabaseMissingInfo


class Neo4jConnectionModel(BaseModel):
    neo4j_url: str
    neo4j_username: str
    neo4j_password: str

    @validator("neo4j_url", "neo4j_username", "neo4j_password")
    def check_string(cls, value):
        if not all(value):
            raise DatabaseMissingInfo(
                "Credential and/or connection URL missing \
                to connect to Neo4J database!"
            )
        return value
